extern crate bitvec;

extern crate hercules_ir;

use std::cell::Cell;
use std::collections::{HashMap, VecDeque};
use std::iter::zip;
use std::mem::{swap, take};

use self::bitvec::prelude::*;

use self::hercules_ir::*;

use crate::*;

pub fn sched_compile(
    module: &Module,
    def_uses: &Vec<ImmutableDefUseMap>,
    typing: &ModuleTyping,
    control_subgraphs: &Vec<Subgraph>,
    fork_join_maps: &Vec<HashMap<NodeID, NodeID>>,
    fork_join_nests: &Vec<HashMap<NodeID, Vec<NodeID>>>,
    antideps: &Vec<Vec<(NodeID, NodeID)>>,
    bbs: &Vec<Vec<NodeID>>,
    plans: &Vec<Plan>,
) -> SModule {
    let stypes = convert_to_sched_ir_types(&module.types);
    let sconstants = convert_to_sched_ir_constants(&module.constants);
    let function_names: HashMap<FunctionID, String> = module
        .functions
        .iter()
        .enumerate()
        .map(|(idx, function)| (FunctionID::new(idx), function.name.clone()))
        .collect();

    let mut functions = HashMap::new();
    let mut manifests = HashMap::new();
    for idx in 0..module.functions.len() {
        let (sfunctions, manifest) = FunctionContext::new(
            &module.functions[idx],
            &module.types,
            &module.constants,
            &module.dynamic_constants,
            &def_uses[idx],
            &typing[idx],
            &control_subgraphs[idx],
            &fork_join_maps[idx],
            &fork_join_nests[idx],
            &antideps[idx],
            &bbs[idx],
            &plans[idx],
            &stypes,
            &sconstants,
            &function_names,
        )
        .compile_function();

        functions.extend(sfunctions.into_iter());
        manifests.insert(module.functions[idx].name.clone(), manifest);
    }

    SModule {
        functions,
        manifests,
    }
}

fn convert_to_sched_ir_types(types: &Vec<Type>) -> Vec<Option<SType>> {
    let mut stypes = vec![None; types.len()];

    for id in types_bottom_up(types) {
        stypes[id.idx()] = match &types[id.idx()] {
            Type::Control => None,
            Type::Boolean => Some(SType::Boolean),
            Type::Integer8 => Some(SType::Integer8),
            Type::Integer16 => Some(SType::Integer16),
            Type::Integer32 => Some(SType::Integer32),
            Type::Integer64 => Some(SType::Integer64),
            Type::UnsignedInteger8 => Some(SType::UnsignedInteger8),
            Type::UnsignedInteger16 => Some(SType::UnsignedInteger16),
            Type::UnsignedInteger32 => Some(SType::UnsignedInteger32),
            Type::UnsignedInteger64 => Some(SType::UnsignedInteger64),
            Type::Float32 => Some(SType::Float32),
            Type::Float64 => Some(SType::Float64),
            Type::Product(fields) => {
                let mut typs = vec![];
                let mut res_none = false;
                for id in fields {
                    if types[id.idx()].is_array() {
                        res_none = true;
                        break;
                    } else {
                        match &stypes[id.idx()] {
                            None => {
                                res_none = true;
                                break;
                            }
                            Some(t) => typs.push(t.clone()),
                        }
                    }
                }
                if res_none {
                    None
                } else {
                    Some(SType::Product(typs.into()))
                }
            }
            Type::Summation(_) => todo!(),
            Type::Array(elem_ty, _) => match &stypes[elem_ty.idx()] {
                None => None,
                Some(t) => Some(SType::ArrayRef(Box::new(t.clone()))),
            },
        };
    }

    stypes
}

fn convert_to_sched_ir_constants(constants: &Vec<Constant>) -> Vec<Option<SConstant>> {
    let mut sconstants = vec![None; constants.len()];

    for id in constants_bottom_up(constants) {
        sconstants[id.idx()] = match &constants[id.idx()] {
            Constant::Boolean(val) => Some(SConstant::Boolean(*val)),
            Constant::Integer8(val) => Some(SConstant::Integer8(*val)),
            Constant::Integer16(val) => Some(SConstant::Integer16(*val)),
            Constant::Integer32(val) => Some(SConstant::Integer32(*val)),
            Constant::Integer64(val) => Some(SConstant::Integer64(*val)),
            Constant::UnsignedInteger8(val) => Some(SConstant::UnsignedInteger8(*val)),
            Constant::UnsignedInteger16(val) => Some(SConstant::UnsignedInteger16(*val)),
            Constant::UnsignedInteger32(val) => Some(SConstant::UnsignedInteger32(*val)),
            Constant::UnsignedInteger64(val) => Some(SConstant::UnsignedInteger64(*val)),
            Constant::Float32(val) => Some(SConstant::Float32(*val)),
            Constant::Float64(val) => Some(SConstant::Float64(*val)),
            Constant::Product(_, fields) => {
                let mut consts = vec![];
                let mut res_none = false;
                for id in fields {
                    if constants[id.idx()].is_array() {
                        res_none = true;
                        break;
                    } else {
                        match &sconstants[id.idx()] {
                            None => {
                                res_none = true;
                                break;
                            }
                            Some(c) => consts.push(c.clone()),
                        }
                    }
                }
                if res_none {
                    None
                } else {
                    Some(SConstant::Product(consts.into()))
                }
            }
            Constant::Summation(_, _, _) => todo!(),
            // Array constants are never generated inline schedule IR.
            Constant::Array(_) => None,
        };
    }

    sconstants
}

/*
 * Converts one Hercules function to N schedule IR functions, where N is the
 * number of partitions in the Hercules function.
 */
struct FunctionContext<'a> {
    function: &'a Function,
    types: &'a Vec<Type>,
    constants: &'a Vec<Constant>,
    dynamic_constants: &'a Vec<DynamicConstant>,
    def_use: &'a ImmutableDefUseMap,
    typing: &'a Vec<TypeID>,
    control_subgraph: &'a Subgraph,
    fork_join_map: &'a HashMap<NodeID, NodeID>,
    fork_join_nest: &'a HashMap<NodeID, Vec<NodeID>>,
    antideps: &'a Vec<(NodeID, NodeID)>,
    bbs: &'a Vec<NodeID>,
    plan: &'a Plan,
    stypes: &'a Vec<Option<SType>>,
    sconstants: &'a Vec<Option<SConstant>>,
    function_names: &'a HashMap<FunctionID, String>,

    top_nodes: Vec<NodeID>,
    partition_graph: Subgraph,
    inverted_partition_map: Vec<Vec<NodeID>>,
    data_inputs: Vec<Vec<NodeID>>,
    data_outputs: Vec<Vec<NodeID>>,

    num_virtual_registers: Vec<Cell<usize>>,
}

impl<'a> FunctionContext<'a> {
    fn new(
        function: &'a Function,
        types: &'a Vec<Type>,
        constants: &'a Vec<Constant>,
        dynamic_constants: &'a Vec<DynamicConstant>,
        def_use: &'a ImmutableDefUseMap,
        typing: &'a Vec<TypeID>,
        control_subgraph: &'a Subgraph,
        fork_join_map: &'a HashMap<NodeID, NodeID>,
        fork_join_nest: &'a HashMap<NodeID, Vec<NodeID>>,
        antideps: &'a Vec<(NodeID, NodeID)>,
        bbs: &'a Vec<NodeID>,
        plan: &'a Plan,
        stypes: &'a Vec<Option<SType>>,
        sconstants: &'a Vec<Option<SConstant>>,
        function_names: &'a HashMap<FunctionID, String>,
    ) -> Self {
        let inverted_partition_map = plan.invert_partition_map();
        let top_nodes = plan.compute_top_nodes(function, control_subgraph, &inverted_partition_map);
        let partition_graph = partition_graph(function, def_use, plan);
        let data_inputs = plan.compute_data_inputs(function);
        let data_outputs = plan.compute_data_outputs(function, def_use);

        let num_virtual_registers = vec![Cell::new(0); plan.num_partitions];

        FunctionContext {
            function,
            types,
            constants,
            dynamic_constants,
            def_use,
            typing,
            control_subgraph,
            fork_join_map,
            fork_join_nest,
            antideps,
            bbs,
            plan,
            stypes,
            sconstants,
            function_names,

            top_nodes,
            partition_graph,
            inverted_partition_map,
            data_inputs,
            data_outputs,

            num_virtual_registers,
        }
    }

    /*
     * Top level function to compile a Hercules IR function into simple IR
     * functions.
     */
    fn compile_function(&self) -> (HashMap<SFunctionName, SFunction>, Manifest) {
        let (mut manifest, array_node_to_array_id) = self.compute_manifest();

        manifest
            .partitions
            .iter()
            .enumerate()
            .for_each(|(idx, partition_manifest)| {
                self.num_virtual_registers[idx].set(partition_manifest.parameters.len())
            });

        let partition_functions = (0..self.plan.num_partitions)
            .map(|partition_idx| {
                let name = self.get_sfunction_name(partition_idx);
                let sfunction =
                    self.compile_partition(partition_idx, &manifest, &array_node_to_array_id);
                self.update_manifest(&mut manifest.partitions[partition_idx], &sfunction);
                (name, sfunction)
            })
            .collect();

        (partition_functions, manifest)
    }

    /*
     * Compute the manifest for a Hercules function. This includes all of the
     * partition signature information.
     */
    fn compute_manifest(&self) -> (Manifest, HashMap<NodeID, ArrayID>) {
        // The manifest needs to contain metadata for allocating arrays.
        let dynamic_constants = self.dynamic_constants.clone();
        let array_constants = self
            .function
            .nodes
            .iter()
            .filter_map(|node| {
                if let Some(cons) = node.try_constant()
                    && let Some(ty) = self.constants[cons.idx()].try_array_type()
                {
                    let extents = self.types[ty.idx()]
                        .try_extents()
                        .expect("PANIC: Type of array constant is not an array type.");
                    Some(extents.into_iter().map(|id| *id).collect())
                } else {
                    None
                }
            })
            .collect();

        // Assign each array constant a unique ID for noting which ones to pass
        // to what partition functions.
        let array_node_to_array_id = (0..self.function.nodes.len())
            .filter(|node_idx| {
                if let Some(cons) = self.function.nodes[*node_idx].try_constant() {
                    self.constants[cons.idx()].is_array()
                } else {
                    false
                }
            })
            .enumerate()
            .map(|(idx, node_idx)| (NodeID::new(node_idx), ArrayID::new(idx)))
            .collect::<HashMap<NodeID, ArrayID>>();

        let partitions = (0..self.plan.num_partitions)
            .map(|partition_idx| {
                let partition = &self.inverted_partition_map[partition_idx];
                let name = self.get_sfunction_name(partition_idx);
                let mut parameters = vec![];
                let mut returns = vec![];

                // Compute the signature of each partitions' schedule IR
                // function, which has the following structure:
                // 1. If the partition is the entry partition, the first
                //    parameters are the parameters to the Hercules function. If
                //    not, then the first parameters are all of the data inputs
                //    to the partition. Note that parameter nodes are always in
                //    the partition of the start node (the entry partition), so
                //    function parameters used in other partitions are treated
                //    as an inter-partition data dependence.
                if partition_idx == 0 {
                    parameters.extend(self.function.param_types.iter().enumerate().map(
                        |(param_idx, ty_id)| {
                            (
                                self.stypes[ty_id.idx()].clone().unwrap(),
                                ParameterKind::HerculesParameter(param_idx),
                            )
                        },
                    ));
                } else {
                    parameters.extend(self.data_inputs[partition_idx].iter().map(|node_id| {
                        (
                            self.stypes[self.typing[node_id.idx()].idx()]
                                .clone()
                                .unwrap(),
                            ParameterKind::DataInput(*node_id),
                        )
                    }))
                }

                // 2. The second set of parameters are references to zero-ed
                //    memories for implementing array constants. Implicit array
                //    cloning is, for now, forbidden. Array constants are
                //    rematerialized into each partition that uses the constant,
                //    so look over all of the uses of all the nodes in the
                //    partition, not all of the nodes in the partition.
                parameters.extend(
                    partition
                        .iter()
                        .map(|node_id| {
                            get_uses(&self.function.nodes[node_id.idx()])
                                .as_ref()
                                .iter()
                                .filter_map(|use_id| {
                                    if let Some(array_id) = array_node_to_array_id.get(use_id) {
                                        Some((
                                            self.stypes[self.typing[use_id.idx()].idx()]
                                                .clone()
                                                .unwrap(),
                                            ParameterKind::ArrayConstant(*array_id),
                                        ))
                                    } else {
                                        None
                                    }
                                })
                                .collect::<Vec<_>>()
                        })
                        .flatten(),
                );

                // 3. The third set of parameters are the dynamic constants
                //    passed to the overall function.
                parameters.extend((0..self.function.num_dynamic_constants).map(|idx| {
                    (
                        SType::UnsignedInteger64,
                        ParameterKind::DynamicConstant(idx as usize),
                    )
                }));

                // Note that many partitions will be given unused parameters
                // (mainly dynamic constants). These will be removed during the
                // small amount of optimization done on simple IR.

                // Simple IR functions may return multiple values (this is to
                // avoid needing to pack / un-pack product types). The return
                // value of an exit partition is the return value of the
                // Hercules function. The return values of non-exit partitions
                // are the data outputs of the partition, possibly plus an
                // integer specifying what partition should be executed next, if
                // there are multiple successor partitions. A valid partitioning
                // will only contain partitions with either a branch to another
                // partition xor a return node.
                let successors = self
                    .partition_graph
                    .succs(NodeID::new(partition_idx))
                    .map(|node_id| PartitionID::new(node_id.idx()))
                    .collect::<Vec<PartitionID>>();
                if partition
                    .iter()
                    .any(|node_id| self.function.nodes[node_id.idx()].is_return())
                {
                    assert_eq!(successors.len(), 0);
                    returns.push((
                        self.stypes[self.function.return_type.idx()]
                            .clone()
                            .unwrap(),
                        ReturnKind::HerculesReturn,
                    ));
                } else {
                    assert!(successors.len() > 0);
                    returns.extend(self.data_outputs[partition_idx].iter().map(|node_id| {
                        (
                            self.stypes[self.typing[node_id.idx()].idx()]
                                .clone()
                                .unwrap(),
                            ReturnKind::DataOutput(*node_id),
                        )
                    }));
                    if successors.len() > 1 {
                        returns.push((SType::Integer8, ReturnKind::NextPartition));
                    }
                }

                let device = match self.plan.partition_devices[partition_idx] {
                    Device::CPU => DeviceManifest::cpu(),
                    Device::GPU => DeviceManifest::gpu(),
                    Device::AsyncRust => todo!(),
                };

                PartitionManifest {
                    name,
                    parameters,
                    returns,
                    successors,
                    device,
                }
            })
            .collect();

        // The parameters for the overall Hercules function is computed in a
        // similar fashion as for the individual partition functions.
        let mut param_types = vec![];
        param_types.extend(self.function.param_types.iter().enumerate().map(
            |(param_idx, ty_id)| {
                (
                    self.stypes[ty_id.idx()].clone().unwrap(),
                    ParameterKind::HerculesParameter(param_idx),
                )
            },
        ));
        param_types.extend(array_node_to_array_id.iter().map(|(node_id, array_id)| {
            (
                self.stypes[self.typing[node_id.idx()].idx()]
                    .clone()
                    .unwrap(),
                ParameterKind::ArrayConstant(*array_id),
            )
        }));
        param_types.extend((0..self.function.num_dynamic_constants).map(|idx| {
            (
                SType::UnsignedInteger64,
                ParameterKind::DynamicConstant(idx as usize),
            )
        }));

        // The return type is just the schedule IR type corresponding to the
        // Hercules function's return type.
        let return_type = self.stypes[self.function.return_type.idx()]
            .clone()
            .unwrap();

        let manifest = Manifest {
            param_types,
            return_type,
            dynamic_constants,
            array_constants,
            partitions,
        };
        (manifest, array_node_to_array_id)
    }

    /*
     * Compile a partition into an SFunction.
     */
    fn compile_partition(
        &self,
        partition_idx: usize,
        manifest: &Manifest,
        array_node_to_array_id: &HashMap<NodeID, ArrayID>,
    ) -> SFunction {
        let partition = &self.inverted_partition_map[partition_idx];
        let mut blocks = vec![];

        // First, create basic blocks inside the SFunction corresponding to the
        // control nodes in the partition. If this isn't the entry partition
        // (partition #0), add an entry block, since the first basic block in a
        // partition may have a predecessor inside the partition.
        let mut control_id_to_block_id = HashMap::new();
        let mut fork_node_id_to_fork_join_id = HashMap::new();
        if partition_idx != 0 {
            // Create an explicit entry block, if one is already created via the
            // Start node.
            blocks.push(SBlock::default());
        }
        for node in partition {
            if self.function.nodes[node.idx()].is_control() {
                control_id_to_block_id.insert(*node, BlockID::new(blocks.len()));
                let mut block = SBlock::default();
                if let Some(imm_fork) = self.fork_join_nest[node].get(0) {
                    let new_id = ForkJoinID::new(fork_node_id_to_fork_join_id.len());
                    let fork_join_id = *fork_node_id_to_fork_join_id
                        .entry(*imm_fork)
                        .or_insert(new_id);
                    block.kind = if self.function.nodes[node.idx()].is_join() {
                        SBlockKind::Reduce(fork_join_id)
                    } else {
                        SBlockKind::Parallel(fork_join_id)
                    };
                }
                blocks.push(block);
            }
        }

        // Second, assign every data node a SValue. This map incorporates info
        // from the manifest to make using SFunction parameters easy.
        let mut data_id_to_svalue = manifest.partitions[partition_idx]
            .parameters
            .iter()
            .enumerate()
            .filter_map(|(idx, (_, kind))| match kind {
                // Assign SValues to nodes defined outside the partition and
                // passed in via SFunction parameters.
                ParameterKind::DataInput(node_id) => Some((*node_id, SValue::VirtualRegister(idx))),
                _ => None,
            })
            .chain(
                // Assign SValues for nodes inside the partition.
                partition
                    .iter()
                    .filter(|node_id| !self.function.nodes[node_id.idx()].is_control())
                    .filter_map(|data_id| {
                        let value = match self.function.nodes[data_id.idx()] {
                            // Phis in a block with no predecessors inside the
                            // current partition don't get lowered to phis in
                            // schedule IR - they get lowered to partition
                            // parameters. Phis with some predecessors in the
                            // same partition and some in a different partition
                            // get lowered to a combination of a SFunction
                            // parameter and a phi instruction, and uses of the
                            // phi node should become uses of the phi
                            // instruction.
                            Node::Phi { control, data: _ }
                                if self.control_subgraph.preds(control).all(|pred| {
                                    self.plan.partitions[pred.idx()]
                                        != self.plan.partitions[control.idx()]
                                }) =>
                            {
                                // If the phi just gets lowered to a parameter,
                                // it got added above when adding the virtual
                                // registers for the SFunction parameters.
                                return None;
                            }
                            // Figure out which virtual constant in the
                            // signature of the current SFunction corresponds to
                            // a particular Hercules parameter.
                            Node::Parameter { index } => SValue::VirtualRegister(
                                manifest.partitions[partition_idx]
                                    .parameters
                                    .iter()
                                    .position(|(_, kind)| {
                                        *kind == ParameterKind::HerculesParameter(index)
                                    })
                                    .unwrap(),
                            ),
                            // Wait to assign SValues to constants. We assign
                            // SValues to constants in user partitions, not in
                            // the partition the constant node happens to be in.
                            Node::Constant { id: _ } => {
                                return None;
                            }
                            // Dynamic constant nodes get generated upfront,
                            // since they may or may not need a virtual register
                            // freshly allocated for them. The math necessary
                            // for them gets put in the block corresponding to
                            // the control node the DynamicConstant node was
                            // scheduled to.
                            Node::DynamicConstant { id } => {
                                let block_id = control_id_to_block_id[&self.bbs[data_id.idx()]];
                                self.compile_dynamic_constant(
                                    id,
                                    &mut blocks[block_id.idx()],
                                    partition_idx,
                                    manifest,
                                )
                            }
                            // Wait to assign SValues to array writes.
                            Node::Write {
                                collect: _,
                                data: _,
                                indices: _,
                            } if self.types[self.typing[data_id.idx()].idx()].is_array() => {
                                return None
                            }
                            _ => SValue::VirtualRegister(self.make_virt_reg(partition_idx)),
                        };
                        Some((*data_id, value))
                    }),
            )
            .chain(
                // Assign SValues for constants used by nodes in the partition.
                partition
                    .iter()
                    .map(|node_id| {
                        get_uses(&self.function.nodes[node_id.idx()])
                            .as_ref()
                            .iter()
                            .filter_map(|use_id| {
                                if let Node::Constant { id } = self.function.nodes[use_id.idx()] {
                                    // Array constants map to the parameter the
                                    // array memory is passed in through - all
                                    // other constants are represented inline in
                                    // an SValue.
                                    let svalue = if let Some(array_id) =
                                        array_node_to_array_id.get(use_id)
                                    {
                                        SValue::VirtualRegister(
                                            manifest.partitions[partition_idx]
                                                .parameters
                                                .iter()
                                                .position(|(_, kind)| {
                                                    *kind == ParameterKind::ArrayConstant(*array_id)
                                                })
                                                .unwrap(),
                                        )
                                    } else {
                                        SValue::Constant(self.sconstants[id.idx()].clone().unwrap())
                                    };
                                    Some((*use_id, svalue))
                                } else {
                                    None
                                }
                            })
                            .collect::<Vec<_>>()
                    })
                    .flatten(),
            )
            .collect::<HashMap<_, _>>();

        // Next, assign all the array write nodes. Array write nodes are
        // recursively assigned the SValue of their `collect` input.
        let mut worklist = partition
            .iter()
            .filter(|id| {
                self.function.nodes[id.idx()].is_write() && !data_id_to_svalue.contains_key(id)
            })
            .map(|id| *id)
            .collect::<VecDeque<_>>();
        while let Some(id) = worklist.pop_front() {
            let pred = match self.function.nodes[id.idx()] {
                Node::Write {
                    data: _,
                    indices: _,
                    collect,
                } => collect,
                _ => panic!("PANIC: Filtered out write nodes, but found a different node kind."),
            };
            if let Some(svalue) = data_id_to_svalue.get(&pred) {
                data_id_to_svalue.insert(id, svalue.clone());
            } else {
                worklist.push_front(id);
            }
        }

        // Third, generate code for every node in the partition. Iterates
        // through a worklist of nodes in the partition. For non-phi and non-
        // reduce nodes, only emit once all data uses are emitted. In addition,
        // consider additional anti-dependence edges from read to write nodes.
        // Def-use and anti-dependence edges are the only ordering we guarantee
        // in schedule IR basic blocks, and it's up to device-specific backends
        // to perform instruction scheduling.
        let mut visited = bitvec![u8, Lsb0; 0; self.function.nodes.len()];
        let mut worklist = partition.iter().map(|id| *id).collect::<VecDeque<_>>();
        while let Some(id) = worklist.pop_front() {
            if self.function.nodes[id.idx()].is_phi()
                || self.function.nodes[id.idx()].is_reduce()
                || get_uses(&self.function.nodes[id.idx()])
                    .as_ref()
                    .into_iter()
                    // If this node isn't a phi or reduce, we need to check that
                    // all uses, as well as all reads we anti-depend with, have
                    // been emitted.
                    .chain(self.antideps.iter().filter_map(|(read, write)| {
                        if id == *write {
                            Some(read)
                        } else {
                            None
                        }
                    }))
                    // Only data dependencies within this partition need to have
                    // already been visited.
                    .all(|use_id| {
                        self.plan.partitions[use_id.idx()] != PartitionID::new(partition_idx)
                            || self.function.nodes[use_id.idx()].is_control()
                            || visited[use_id.idx()]
                    })
            {
                // Once all of the data dependencies for this node are emitted,
                // this node can be emitted.
                self.compile_node(
                    id,
                    &control_id_to_block_id,
                    &data_id_to_svalue,
                    &fork_node_id_to_fork_join_id,
                    &mut blocks,
                    partition_idx,
                    manifest,
                );
                visited.set(id.idx(), true);
            } else {
                // Skip emitting node if it's not a phi or reduce node and if
                // its data uses are not emitted yet.
                worklist.push_back(id);
            }
        }

        // Fourth, add the jump from the explicit entry block to the top node's
        // block in the partition.
        if partition_idx != 0 {
            // Explicitly jump to the block corresponding to the top of the
            // partition. That block may be a parallel block, but it's not a
            // reduce block.
            let top_node = self.top_nodes[partition_idx];
            let top_block = control_id_to_block_id[&top_node];
            let parallel_entry = if self.function.nodes[top_node.idx()].is_fork() {
                self.copy_schedules(top_node, &mut blocks[0]);
                Some(self.compile_parallel_entry(
                    top_node,
                    &data_id_to_svalue,
                    &mut blocks[0],
                    partition_idx,
                    manifest,
                ))
            } else {
                None
            };
            blocks[0].insts.push(SInst::Jump {
                target: top_block,
                parallel_entry,
                reduce_exit: None,
            });
            blocks[0]
                .virt_regs
                .push((self.make_virt_reg(partition_idx), SType::Boolean));
        }

        // Fifth, make sure every block's schedules map is "filled".
        for block in blocks.iter_mut() {
            for inst_idx in 0..block.insts.len() {
                let _ = block.schedules.try_insert(inst_idx, vec![]);
            }
        }

        SFunction {
            blocks,
            param_types: manifest.partitions[partition_idx]
                .parameters
                .iter()
                .map(|(sty, _)| sty.clone())
                .collect(),
            return_types: manifest.partitions[partition_idx]
                .returns
                .iter()
                .map(|(sty, _)| sty.clone())
                .collect(),
        }
    }

    fn compile_node(
        &self,
        id: NodeID,
        control_id_to_block_id: &HashMap<NodeID, BlockID>,
        data_id_to_svalue: &HashMap<NodeID, SValue>,
        fork_node_id_to_fork_join_id: &HashMap<NodeID, ForkJoinID>,
        blocks: &mut Vec<SBlock>,
        partition_idx: usize,
        manifest: &Manifest,
    ) {
        let bb = self.bbs[id.idx()];
        let block_id = Cell::new(control_id_to_block_id[&bb]);
        let mut block = take(&mut blocks[block_id.get().idx()]);

        // Uses of reduce nodes inside their corresponding reduce block need to
        // refer to the reduction variable instruction, not the output of the
        // reduce block.
        let get_svalue = |id: NodeID| data_id_to_svalue[&id].clone();
        let self_virt_reg = || get_svalue(id).try_virt_reg().unwrap();

        // Helper function to lower a jump to a particular control node.
        let lower_jmp = |dst: NodeID, block: &mut SBlock| {
            if let Some(block_id) = control_id_to_block_id.get(&dst) {
                // The successor block is in this partition. Add extra info to
                // the jump if we're jumping into a parallel section or out of a
                // reduce section. Note that both of those may be true at once.
                let parallel_entry = if self.function.nodes[dst.idx()].is_fork() {
                    self.copy_schedules(dst, block);
                    Some(self.compile_parallel_entry(
                        dst,
                        data_id_to_svalue,
                        block,
                        partition_idx,
                        manifest,
                    ))
                } else {
                    None
                };
                let reduce_exit = if self.function.nodes[id.idx()].is_join() {
                    Some(self.compile_reduce_exit(id, data_id_to_svalue))
                } else {
                    None
                };
                block.insts.push(SInst::Jump {
                    target: *block_id,
                    parallel_entry,
                    reduce_exit,
                });
            } else {
                assert_ne!(
                    self.plan.partitions[id.idx()],
                    self.plan.partitions[dst.idx()]
                );

                // The successor block is in a different partition.
                let next_partition = self.plan.partitions[dst.idx()];
                let data_outputs = manifest.partitions[partition_idx]
                    .returns
                    .iter()
                    .map(|(_, kind)| match kind {
                        ReturnKind::DataOutput(id) => get_svalue(*id).clone(),
                        ReturnKind::HerculesReturn => panic!("PANIC: Partition can't contain a HerculesReturn kind of return value when it jumps to another partition."),
                        ReturnKind::NextPartition => SValue::Constant(SConstant::Integer8(next_partition.idx() as i8)),
                    })
                    .collect();

                block.insts.push(SInst::PartitionExit { data_outputs });
            }
            block
                .virt_regs
                .push((self.make_virt_reg(partition_idx), SType::Boolean));
        };

        // Helper function to generate the dynamic constant math to compute the
        // bounds of an array type of a node.
        let lower_extents = |id: NodeID, block: &mut SBlock| {
            self.types[self.typing[id.idx()].idx()]
                .try_extents()
                .unwrap()
                .iter()
                .map(|dc| self.compile_dynamic_constant(*dc, block, partition_idx, manifest))
                .collect()
        };

        // Emit schedule IR instructions corresponding to this Hercules IR node.
        match self.function.nodes[id.idx()] {
            // Forks are super simple to lower here. Since what's sequential /
            // parallel / reducing is encoded in basic block kinds, and entry /
            // exits are handled in `lower_jmp`, we just need to add a jump like
            // any other control block with one successor.
            Node::Start
            | Node::Region { preds: _ }
            | Node::Projection {
                control: _,
                selection: _,
            }
            | Node::Fork {
                control: _,
                factors: _,
            } => {
                let mut succs = self.control_subgraph.succs(id);
                assert_eq!(succs.len(), 1);
                let succ = succs.next().unwrap();
                lower_jmp(succ, &mut block);
            }
            Node::Join { control: _ } => {
                let mut succs = self.control_subgraph.succs(id);
                assert_eq!(succs.len(), 1);
                let succ = succs.next().unwrap();
                if self.plan.partitions[id.idx()] != self.plan.partitions[succ.idx()] {
                    // If the successor is in another partition, we need to add
                    // a sequential block to hold the PartitionExit. Add a jump,
                    // with reduce exit metadata, to the reduce block.
                    let exit_block_id = BlockID::new(blocks.len());
                    let reduce_exit = self.compile_reduce_exit(id, data_id_to_svalue);
                    block.insts.push(SInst::Jump {
                        target: exit_block_id,
                        parallel_entry: None,
                        reduce_exit: Some(reduce_exit),
                    });
                    block
                        .virt_regs
                        .push((self.make_virt_reg(partition_idx), SType::Boolean));
                    // The exit block contains just a PartitionExit instruction.
                    let mut exit_block = SBlock::default();
                    // `lower_jmp` depends on `block_id`, so temporarily update.
                    let old_block_id = block_id.get();
                    block_id.set(exit_block_id);
                    lower_jmp(succ, &mut exit_block);
                    block_id.set(old_block_id);
                    blocks.push(exit_block);
                } else {
                    // Otherwise, lower the jump as normal.
                    lower_jmp(succ, &mut block);
                }
            }
            Node::If { control: _, cond } => {
                let mut succs = self.control_subgraph.succs(id);
                let mut proj1 = succs.next().unwrap();
                let mut proj2 = succs.next().unwrap();
                assert_eq!(succs.next(), None);
                if self.function.nodes[proj1.idx()].try_proj().unwrap().1 == 1 {
                    swap(&mut proj1, &mut proj2);
                }
                block.insts.push(SInst::Branch {
                    cond: get_svalue(cond).clone(),
                    false_target: control_id_to_block_id[&self.bbs[proj1.idx()]],
                    true_target: control_id_to_block_id[&self.bbs[proj2.idx()]],
                });
                block
                    .virt_regs
                    .push((self.make_virt_reg(partition_idx), SType::Boolean));
            }
            Node::Return { control: _, data } => {
                block.insts.push(SInst::Return {
                    value: get_svalue(data).clone(),
                });
                block
                    .virt_regs
                    .push((self.make_virt_reg(partition_idx), SType::Boolean));
            }

            Node::Phi { control, ref data } => {
                let control_uses = get_uses(&self.function.nodes[control.idx()]);
                let mut found_in_partition_predecessor = false;
                let mut found_out_of_partition_predecessor = false;
                let inputs = zip(control_uses.as_ref().iter(), data.iter())
                    .filter_map(|(control_use, data_id)| {
                        if let Some(block_id) = control_id_to_block_id.get(control_use) {
                            // If any of the predecessors are in this partition,
                            // we actually generate a phi instruction.
                            // Otherwise, we just need to refer to the parameter
                            // of the SFunction corresponding to this phi.
                            found_in_partition_predecessor = true;
                            Some((*block_id, get_svalue(*data_id).clone()))
                        } else if let Some(param_idx) = manifest.partitions[partition_idx]
                            .parameters
                            .iter()
                            .position(|(_, kind)| *kind == ParameterKind::DataInput(id))
                        {
                            // This input to the phi is corresponds to all of
                            // the inputs from control locations outside this
                            // partition. This does *not* include constant nodes
                            // in other partitions - those get propagated (see
                            // below). Don't add multiple inputs for block #0.
                            if found_out_of_partition_predecessor {
                                return None;
                            }
                            // This predecessor for the phi gets passed in
                            // via a parameter set up for this phi.
                            found_out_of_partition_predecessor = true;
                            Some((BlockID::new(0), SValue::VirtualRegister(param_idx)))
                        } else {
                            // This input to the phi is a constant located
                            // outside this partition these get propagated in
                            // schedule IR.
                            found_in_partition_predecessor = true;
                            let svalue = get_svalue(*data_id).clone();
                            Some((BlockID::new(0), svalue))
                        }
                    })
                    .collect();

                // If there's at least one predecessor inside this partition, we
                // need to generate an actual phi instruction.
                if found_in_partition_predecessor {
                    block.insts.push(SInst::Phi { inputs });
                    block.virt_regs.push((
                        self_virt_reg(),
                        self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                    ));
                }
            }

            Node::ThreadID { control, dimension } => {
                let fork_join = fork_node_id_to_fork_join_id[&control];
                block.insts.push(SInst::ThreadID {
                    dimension,
                    fork_join,
                });
                block
                    .virt_regs
                    .push((self_virt_reg(), SType::UnsignedInteger64));
            }
            Node::Reduce {
                control,
                init: _,
                reduct: _,
            } => {
                // Determine the reduction variable number based on the users of
                // the join node.
                let number = self
                    .def_use
                    .get_users(control)
                    .iter()
                    .filter(|user| self.function.nodes[user.idx()].is_reduce())
                    .position(|user| *user == id)
                    .unwrap();
                self.copy_schedules(id, &mut block);
                block.insts.push(SInst::ReductionVariable { number });
                block.virt_regs.push((
                    self_virt_reg(),
                    self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                ));
            }

            Node::Unary { input, op } => {
                block.insts.push(SInst::Unary {
                    input: get_svalue(input).clone(),
                    op: convert_unary_op(op, &self.stypes),
                });
                block.virt_regs.push((
                    self_virt_reg(),
                    self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                ));
            }
            Node::Binary { left, right, op } => {
                block.insts.push(SInst::Binary {
                    left: get_svalue(left).clone(),
                    right: get_svalue(right).clone(),
                    op: convert_binary_op(op),
                });
                block.virt_regs.push((
                    self_virt_reg(),
                    self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                ));
            }
            Node::Ternary {
                first,
                second,
                third,
                op,
            } => {
                block.insts.push(SInst::Ternary {
                    first: get_svalue(first).clone(),
                    second: get_svalue(second).clone(),
                    third: get_svalue(third).clone(),
                    op: convert_ternary_op(op),
                });
                block.virt_regs.push((
                    self_virt_reg(),
                    self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                ));
            }

            Node::Read {
                collect,
                ref indices,
            } => {
                let mut collect_svalue = get_svalue(collect);
                let mut prod_indices = &indices[..];

                // We currently only support top-level arrays and products. The
                // array and product portions become separate instructions.
                // Since arrays are always root types, handle them first.
                if let Some(position) = indices[0].try_position() {
                    // If there's both an array load and a product extract, we
                    // need to allocate an intermediary virtual register.
                    let dst_virt_reg = if indices.len() > 1 {
                        self.make_virt_reg(partition_idx)
                    } else {
                        self_virt_reg()
                    };

                    let position = position.iter().map(|id| get_svalue(*id)).collect();
                    // Array loads need the dynamic constant bounds for indexing
                    // math.
                    let bounds = lower_extents(collect, &mut block);
                    let load_ty = if let SType::ArrayRef(elem_ty) = self.stypes
                        [self.typing[collect.idx()].idx()]
                    .clone()
                    .unwrap()
                    {
                        *elem_ty
                    } else {
                        panic!("PANIC: Type of collection isn't an array when an ArrayLoad use is generated.")
                    };
                    block.insts.push(SInst::ArrayLoad {
                        array: collect_svalue,
                        position,
                        bounds,
                    });
                    block.virt_regs.push((dst_virt_reg, load_ty));

                    // The product extract needs to extract from the product
                    // loaded from the array.
                    collect_svalue = SValue::VirtualRegister(dst_virt_reg);
                    prod_indices = &indices[1..];
                }

                // Handle the product indices.
                if prod_indices.len() > 0 {
                    let indices = prod_indices
                        .iter()
                        .map(|index| index.try_field().unwrap())
                        .collect();
                    block.insts.push(SInst::ProductExtract {
                        product: collect_svalue,
                        indices,
                    });
                    block.virt_regs.push((
                        self_virt_reg(),
                        self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                    ));
                }
            }
            Node::Write {
                collect,
                data,
                ref indices,
            } => {
                // We currently only support top-level arrays and products.
                // There are three cases that we handle separately:
                // 1. Writing to an array. This just lowers to an ArrayStore.
                // 2. Writing to a product inside an array. This lowers to an
                //    ArrayLoad to get the initial product value, a
                //    ProductInsert to update the product value, and an
                //    ArrayStore to write the new product value into the array.
                // 3. Writing to a product. This just lowers to a ProductInsert.

                if let Some(position) = indices[0].try_position()
                    && indices.len() == 1
                {
                    // Handle case #1.
                    let position = position.iter().map(|id| get_svalue(*id)).collect();
                    // Array stores need the dynamic constant bounds for
                    // indexing math.
                    let bounds = lower_extents(collect, &mut block);
                    block.insts.push(SInst::ArrayStore {
                        array: get_svalue(collect),
                        value: get_svalue(data),
                        position,
                        bounds,
                    });
                    // Array stores don't produce a meaningful virtual register.
                    block
                        .virt_regs
                        .push((self.make_virt_reg(partition_idx), SType::Boolean));
                } else if let Some(position) = indices[0].try_position() {
                    // Handle case #2.
                    let position = position
                        .iter()
                        .map(|id| get_svalue(*id))
                        .collect::<Box<[_]>>();
                    let bounds = lower_extents(collect, &mut block);

                    // Load the product.
                    let load_virt_reg = self.make_virt_reg(partition_idx);
                    let load_ty = if let SType::ArrayRef(elem_ty) = self.stypes
                        [self.typing[collect.idx()].idx()]
                    .clone()
                    .unwrap()
                    {
                        *elem_ty
                    } else {
                        panic!("PANIC: Type of collection isn't an array when an ArrayLoad use is generated.")
                    };
                    block.insts.push(SInst::ArrayLoad {
                        array: get_svalue(collect),
                        position: position.clone(),
                        bounds: bounds.clone(),
                    });
                    block.virt_regs.push((load_virt_reg, load_ty.clone()));

                    // Update the product.
                    let update_virt_reg = self.make_virt_reg(partition_idx);
                    let indices = indices[1..]
                        .iter()
                        .map(|index| index.try_field().unwrap())
                        .collect();
                    block.insts.push(SInst::ProductInsert {
                        product: SValue::VirtualRegister(load_virt_reg),
                        data: get_svalue(data),
                        indices,
                    });
                    block.virt_regs.push((update_virt_reg, load_ty));

                    // Store the product.
                    block.insts.push(SInst::ArrayStore {
                        array: get_svalue(collect),
                        value: SValue::VirtualRegister(update_virt_reg),
                        position,
                        bounds,
                    });
                    block
                        .virt_regs
                        .push((self.make_virt_reg(partition_idx), SType::Boolean));
                } else {
                    // Handle case #3.
                    let indices = indices
                        .iter()
                        .map(|index| index.try_field().unwrap())
                        .collect();
                    block.insts.push(SInst::ProductInsert {
                        product: get_svalue(collect),
                        data: get_svalue(data),
                        indices,
                    });
                    // Product insertions do produce a virtual register, since
                    // they create a new product value.
                    block.virt_regs.push((
                        self_virt_reg(),
                        self.stypes[self.typing[id.idx()].idx()].clone().unwrap(),
                    ));
                }
            }

            // There are a few nodes for which no code needs to get emitted.
            _ => {}
        }

        blocks[block_id.get().idx()] = block;
    }

    /*
     * Helper to copy over schedules.
     */
    fn copy_schedules(&self, src: NodeID, block: &mut SBlock) {
        block.schedules.insert(
            block.insts.len(),
            self.plan.schedules[src.idx()]
                .iter()
                .map(|schedule| sched_make_schedule(schedule))
                .collect(),
        );
    }

    /*
     * Compiles a reference to a dynamic constant into math to compute that
     * dynamic constant. We need a mutable reference to some basic block, since
     * we may need to generate math inline to compute the dynamic constant.
     */
    fn compile_dynamic_constant(
        &self,
        dc: DynamicConstantID,
        block: &mut SBlock,
        partition_idx: usize,
        manifest: &Manifest,
    ) -> SValue {
        match self.dynamic_constants[dc.idx()] {
            DynamicConstant::Constant(cons) => {
                SValue::Constant(SConstant::UnsignedInteger64(cons as u64))
            }
            DynamicConstant::Parameter(idx) => SValue::VirtualRegister(
                manifest.partitions[partition_idx]
                    .parameters
                    .iter()
                    .position(|(_, kind)| *kind == ParameterKind::DynamicConstant(idx))
                    .unwrap(),
            ),

            DynamicConstant::Add(left, right)
            | DynamicConstant::Sub(left, right)
            | DynamicConstant::Mul(left, right)
            | DynamicConstant::Div(left, right)
            | DynamicConstant::Rem(left, right) => {
                let left = self.compile_dynamic_constant(left, block, partition_idx, manifest);
                let right = self.compile_dynamic_constant(right, block, partition_idx, manifest);
                let output_virt_reg = self.make_virt_reg(partition_idx);
                block.insts.push(SInst::Binary {
                    left,
                    right,
                    op: match self.dynamic_constants[dc.idx()] {
                        DynamicConstant::Add(_, _) => SBinaryOperator::Add,
                        DynamicConstant::Sub(_, _) => SBinaryOperator::Sub,
                        DynamicConstant::Mul(_, _) => SBinaryOperator::Mul,
                        DynamicConstant::Div(_, _) => SBinaryOperator::Div,
                        DynamicConstant::Rem(_, _) => SBinaryOperator::Rem,
                        _ => panic!(),
                    },
                });
                block
                    .virt_regs
                    .push((output_virt_reg, SType::UnsignedInteger64));
                SValue::VirtualRegister(output_virt_reg)
            }
        }
    }

    /*
     * Makes a parallel entry for a jump to a fork.
     */
    fn compile_parallel_entry(
        &self,
        fork: NodeID,
        data_id_to_svalue: &HashMap<NodeID, SValue>,
        block: &mut SBlock,
        partition_idx: usize,
        manifest: &Manifest,
    ) -> ParallelEntry {
        let (_, factors) = self.function.nodes[fork.idx()].try_fork().unwrap();
        let thread_counts = factors
            .iter()
            .map(|dc_id| self.compile_dynamic_constant(*dc_id, block, partition_idx, manifest))
            .collect();
        let reduce_inits = self
            .def_use
            .get_users(self.fork_join_map[&fork])
            .iter()
            .filter_map(|user| self.function.nodes[user.idx()].try_reduce())
            .map(|(_, init, _)| data_id_to_svalue[&init].clone())
            .collect();
        ParallelEntry {
            thread_counts,
            reduce_inits,
        }
    }

    /*
     * Makes a reduce exit for a jump from a join.
     */
    fn compile_reduce_exit(
        &self,
        join: NodeID,
        data_id_to_svalue: &HashMap<NodeID, SValue>,
    ) -> ReduceExit {
        let reduce_reducts = self
            .def_use
            .get_users(join)
            .iter()
            .filter(|user| self.function.nodes[user.idx()].is_reduce())
            .map(|reduce| {
                // The SValues that get passed to the reduce exit are the
                // `reduct` input to the reduce node.
                data_id_to_svalue[&get_uses(&self.function.nodes[reduce.idx()]).as_ref()[2]].clone()
            })
            .collect();
        ReduceExit { reduce_reducts }
    }

    fn make_virt_reg(&self, partition_idx: usize) -> usize {
        let virt_reg = self.num_virtual_registers[partition_idx].get();
        self.num_virtual_registers[partition_idx].set(virt_reg + 1);
        virt_reg
    }

    fn get_sfunction_name(&self, partition_idx: usize) -> SFunctionName {
        format!("{}_{}", self.function.name, partition_idx)
    }

    /*
     * There is some information we can only add to the manifest once we've
     * computed the schedule IR.
     */
    fn update_manifest(&self, manifest: &mut PartitionManifest, function: &SFunction) {
        let parallel_reduce_infos = sched_parallel_reduce_sections(function);

        // Add parallel launch info for CPU partitions. This relies on checking
        // schedules inside the generated schedule IR.
        let partition_name = manifest.name.clone();
        if let Some(tiles) = function.blocks[0].schedules[&0]
            .iter()
            .filter_map(|schedule| schedule.try_parallel_launch())
            .next()
            && parallel_reduce_infos
                .into_iter()
                .any(|(_, info)| info.top_level)
            && let DeviceManifest::CPU { parallel_launch } = &mut manifest.device
        {
            let parallel_entry = function.blocks[0].insts[0].try_jump().unwrap().1.unwrap();
            assert_eq!(tiles.len(), parallel_entry.thread_counts.len());
            let top_level_fork_id = self
                .fork_join_nest
                .iter()
                // Find control nodes in the fork join nesting whose only nest
                // it itself (is a top level fork-join).
                .filter(|(id, nest)| nest.len() == 1 && nest[0] == **id)
                // Only consider forks in this partition.
                .filter(|(id, _)| {
                    self.get_sfunction_name(self.plan.partitions[id.idx()].idx()) == partition_name
                })
                .next()
                .unwrap()
                .0;
            *parallel_launch = zip(
                tiles.into_iter(),
                self.function.nodes[top_level_fork_id.idx()]
                    .try_fork()
                    .unwrap()
                    .1,
            )
            .map(|(num_chunks, count_dc_id)| (*num_chunks, *count_dc_id))
            .collect();
        }
    }
}

fn convert_unary_op(op: UnaryOperator, simple_ir_types: &[Option<SType>]) -> SUnaryOperator {
    match op {
        UnaryOperator::Not => SUnaryOperator::Not,
        UnaryOperator::Neg => SUnaryOperator::Neg,
        UnaryOperator::Cast(ty) => SUnaryOperator::Cast(simple_ir_types[ty.idx()].clone().unwrap()),
    }
}

fn convert_binary_op(op: BinaryOperator) -> SBinaryOperator {
    match op {
        BinaryOperator::Add => SBinaryOperator::Add,
        BinaryOperator::Sub => SBinaryOperator::Sub,
        BinaryOperator::Mul => SBinaryOperator::Mul,
        BinaryOperator::Div => SBinaryOperator::Div,
        BinaryOperator::Rem => SBinaryOperator::Rem,
        BinaryOperator::LT => SBinaryOperator::LT,
        BinaryOperator::LTE => SBinaryOperator::LTE,
        BinaryOperator::GT => SBinaryOperator::GT,
        BinaryOperator::GTE => SBinaryOperator::GTE,
        BinaryOperator::EQ => SBinaryOperator::EQ,
        BinaryOperator::NE => SBinaryOperator::NE,
        BinaryOperator::Or => SBinaryOperator::Or,
        BinaryOperator::And => SBinaryOperator::And,
        BinaryOperator::Xor => SBinaryOperator::Xor,
        BinaryOperator::LSh => SBinaryOperator::LSh,
        BinaryOperator::RSh => SBinaryOperator::RSh,
    }
}

fn convert_ternary_op(op: TernaryOperator) -> STernaryOperator {
    match op {
        TernaryOperator::Select => STernaryOperator::Select,
    }
}
