extern crate ordered_float;
extern crate serde;

extern crate hercules_ir;

use std::collections::HashMap;

use self::serde::Deserialize;
use self::serde::Serialize;

use self::hercules_ir::*;

use crate::*;

/*
 * A schedule IR module is a list of functions and a description of each
 * Hercules function in terms of schedule IR functions (called the manifest).
 */
#[derive(Debug, Default, Clone)]
pub struct SModule {
    // Refer to schedule IR functions by their name.
    pub functions: HashMap<SFunctionName, SFunction>,
    // Each Hercules function maps to a manifest.
    pub manifests: HashMap<String, Manifest>,
}

/*
 * A schedule IR function consists of a CFG of basic blocks, each containing
 * instructions. Instructions can produce virtual register outputs, and SSA form
 * is guaranteed. SFunctions can have multiple parameters and return values -
 * many values may cross partition boundaries at once.
 *
 * Since SFunctions represent partitions, many SFunctions don't "return".
 * Instead, conceptually they "jump" to the next partition to run. SFunctions
 * that jump to another partition contain the "PartitionExit" instruction, while
 * SFunctions that return from the Hercules function contain the "Return"
 * instruction. An SFunction must contain either PartitionExits xor Returns.
 *
 * There are two special kinds of basic blocks for representing fork-joins:
 * parallel blocks and reduce blocks. Each parallel / reduce block is associated
 * with a unique ID per fork-join. A parallel block can contain a "ThreadID"
 * instruction, which gets the Nth thread ID. A reduce block can contain a
 * "Reduce" instruction, which gets the last value of the Mth reduction
 * variable. When jumping to a parallel block, a u64 must be provided per fork
 * dimension, specifying how many threads should spawn, and an initial value per
 * reduction variable must be provided. When jumping out of a reduce block, a
 * "new" value for each reduction variable must be provided.
 */
#[derive(Debug, Default, Clone)]
pub struct SFunction {
    pub blocks: Vec<SBlock>,
    pub param_types: Vec<SType>,
    pub return_types: Vec<SType>,
}

impl SFunction {
    pub fn get_inst(&self, id: InstID) -> &SInst {
        &self.blocks[id.idx_0()].insts[id.idx_1()]
    }

    pub fn get_inst_mut(&mut self, id: InstID) -> &mut SInst {
        &mut self.blocks[id.idx_0()].insts[id.idx_1()]
    }
}

/*
 * Use a very simple representation for blocks, since modification is not a
 * priority. Unlike many IRs (say LLVM), the instructions in schedule IR blocks
 * aren't necessarily ordered, as different backends may have different
 * scheduling considerations. This means that, for example, each SBlock must
 * contain exactly one terminating instruction, but the position of that
 * instruction may not be at the end of the block. All that's required is that
 * defs precede uses, and that loads and stores to array references are ordered.
 */
#[derive(Debug, Default, Clone)]
pub struct SBlock {
    pub insts: Vec<SInst>,
    // The virtual registers created by each instruction. Technically, this will
    // assign instructions like ArrayStores and Regions a virtual register,
    // which doesn't make sense. These virtual registers are just ignored. Each
    // virtual register has a certain schedule IR type. The type of virtual
    // registers produced by certain instructions, like Jump or ArrayStore, is
    // set to SType::Boolean, but it's not meaningful.
    pub virt_regs: Vec<(usize, SType)>,
    // Map from instruction index in the block to a list of schedules attached
    // to that instruction.
    pub schedules: HashMap<usize, Vec<SSchedule>>,
    pub kind: SBlockKind,
}

impl SBlock {
    pub fn successors(&self) -> BlockSuccessors {
        self.insts
            .iter()
            .map(|inst| inst.block_successors())
            .filter(|successors| *successors != BlockSuccessors::Zero)
            .next()
            .unwrap_or(BlockSuccessors::Zero)
    }
}

#[derive(Debug, Default, Clone, PartialEq, Eq)]
pub enum SBlockKind {
    #[default]
    Sequential,
    Parallel(ForkJoinID),
    Reduce(ForkJoinID),
}

impl SBlockKind {
    pub fn try_parallel(&self) -> Option<ForkJoinID> {
        if let SBlockKind::Parallel(id) = self {
            Some(*id)
        } else {
            None
        }
    }

    pub fn try_reduce(&self) -> Option<ForkJoinID> {
        if let SBlockKind::Reduce(id) = self {
            Some(*id)
        } else {
            None
        }
    }

    pub fn try_fork_join_id(&self) -> Option<ForkJoinID> {
        match self {
            SBlockKind::Sequential => None,
            SBlockKind::Parallel(id) | SBlockKind::Reduce(id) => Some(*id),
        }
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum SSchedule {
    ParallelLaunch(Box<[usize]>),
    ParallelReduce,
    Vectorizable(usize),
    Associative,
}

impl SSchedule {
    pub fn try_parallel_launch(&self) -> Option<&[usize]> {
        if let SSchedule::ParallelLaunch(tiles) = self {
            Some(tiles)
        } else {
            None
        }
    }

    pub fn try_vectorizable(&self) -> Option<usize> {
        if let SSchedule::Vectorizable(width) = self {
            Some(*width)
        } else {
            None
        }
    }
}

pub fn sched_make_schedule(schedule: &Schedule) -> SSchedule {
    match schedule {
        Schedule::ParallelFork(tiles) => SSchedule::ParallelLaunch(tiles.clone()),
        Schedule::ParallelReduce => SSchedule::ParallelReduce,
        Schedule::Vectorizable(width) => SSchedule::Vectorizable(*width),
        Schedule::Associative => SSchedule::Associative,
    }
}

/*
 * Unlike Hercules IR, we can represent a reference to an array (so that we
 * don't need to use an array value in this IR). This is fine, since we're not
 * doing much analysis / optimization at this stage, and most platforms we want
 * to target have a similar model for working with arrays anyway. We still need
 * value product types, since the layout of these types may be platform
 * dependent.
 */
#[derive(Debug, Clone, PartialOrd, Ord, PartialEq, Eq, Hash, Serialize, Deserialize)]
pub enum SType {
    Boolean,
    Integer8,
    Integer16,
    Integer32,
    Integer64,
    UnsignedInteger8,
    UnsignedInteger16,
    UnsignedInteger32,
    UnsignedInteger64,
    Float32,
    Float64,
    // Don't intern STypes to make developing the code generator easier.
    Product(Box<[SType]>),
    // Array types don't include their bounds, since dynamic constants are not
    // an IR-level concept in schedule IR.
    ArrayRef(Box<SType>),
}

impl SType {
    pub fn is_float(&self) -> bool {
        match self {
            SType::Float32 | SType::Float64 => true,
            _ => false,
        }
    }

    pub fn is_unsigned(&self) -> bool {
        match self {
            SType::UnsignedInteger8
            | SType::UnsignedInteger16
            | SType::UnsignedInteger32
            | SType::UnsignedInteger64 => true,
            _ => false,
        }
    }

    pub fn try_product(&self) -> Option<&[SType]> {
        if let SType::Product(fields) = self {
            Some(fields)
        } else {
            None
        }
    }
}

/*
 * Represents constants, except for array constants.
 */
#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum SConstant {
    Boolean(bool),
    Integer8(i8),
    Integer16(i16),
    Integer32(i32),
    Integer64(i64),
    UnsignedInteger8(u8),
    UnsignedInteger16(u16),
    UnsignedInteger32(u32),
    UnsignedInteger64(u64),
    Float32(ordered_float::OrderedFloat<f32>),
    Float64(ordered_float::OrderedFloat<f64>),
    // Don't intern SConstants to make developing the code generator easier.
    Product(Box<[SConstant]>),
}

impl SConstant {
    pub fn get_type(&self) -> SType {
        match self {
            SConstant::Boolean(_) => SType::Boolean,
            SConstant::Integer8(_) => SType::Integer8,
            SConstant::Integer16(_) => SType::Integer16,
            SConstant::Integer32(_) => SType::Integer32,
            SConstant::Integer64(_) => SType::Integer64,
            SConstant::UnsignedInteger8(_) => SType::UnsignedInteger8,
            SConstant::UnsignedInteger16(_) => SType::UnsignedInteger16,
            SConstant::UnsignedInteger32(_) => SType::UnsignedInteger32,
            SConstant::UnsignedInteger64(_) => SType::UnsignedInteger64,
            SConstant::Float32(_) => SType::Float32,
            SConstant::Float64(_) => SType::Float64,
            SConstant::Product(fields) => {
                SType::Product(fields.into_iter().map(|field| field.get_type()).collect())
            }
        }
    }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum SValue {
    Constant(SConstant),
    VirtualRegister(usize),
}

impl SValue {
    pub fn try_virt_reg(&self) -> Option<usize> {
        if let SValue::VirtualRegister(vr) = self {
            Some(*vr)
        } else {
            None
        }
    }

    pub fn try_constant(&self) -> Option<&SConstant> {
        if let SValue::Constant(cons) = self {
            Some(cons)
        } else {
            None
        }
    }
}

/*
 * Typical instructions of a CFG + SSA IR, plus some instructions for
 * representing particular Hercules IR quirks.
 */
#[derive(Debug, Clone)]
pub enum SInst {
    Phi {
        inputs: Box<[(BlockID, SValue)]>,
    },
    ThreadID {
        dimension: usize,
        fork_join: ForkJoinID,
    },
    ReductionVariable {
        number: usize,
    },
    Jump {
        target: BlockID,
        parallel_entry: Option<ParallelEntry>,
        reduce_exit: Option<ReduceExit>,
    },
    Branch {
        cond: SValue,
        false_target: BlockID,
        true_target: BlockID,
    },
    PartitionExit {
        data_outputs: Box<[SValue]>,
    },
    Return {
        value: SValue,
    },
    Unary {
        input: SValue,
        op: SUnaryOperator,
    },
    Binary {
        left: SValue,
        right: SValue,
        op: SBinaryOperator,
    },
    Ternary {
        first: SValue,
        second: SValue,
        third: SValue,
        op: STernaryOperator,
    },
    ProductExtract {
        product: SValue,
        indices: Box<[usize]>,
    },
    ProductInsert {
        product: SValue,
        data: SValue,
        indices: Box<[usize]>,
    },
    ArrayLoad {
        array: SValue,
        position: Box<[SValue]>,
        bounds: Box<[SValue]>,
    },
    ArrayStore {
        array: SValue,
        value: SValue,
        position: Box<[SValue]>,
        bounds: Box<[SValue]>,
    },
}

impl SInst {
    pub fn is_reduction_variable(&self) -> bool {
        if let SInst::ReductionVariable { number: _ } = self {
            true
        } else {
            false
        }
    }

    pub fn is_phi(&self) -> bool {
        if let SInst::Phi { inputs: _ } = self {
            true
        } else {
            false
        }
    }

    pub fn is_jump(&self) -> bool {
        if let SInst::Jump {
            target: _,
            parallel_entry: _,
            reduce_exit: _,
        } = self
        {
            true
        } else {
            false
        }
    }

    pub fn is_partition_exit(&self) -> bool {
        if let SInst::PartitionExit { data_outputs: _ } = self {
            true
        } else {
            false
        }
    }

    pub fn is_return(&self) -> bool {
        if let SInst::Return { value: _ } = self {
            true
        } else {
            false
        }
    }

    pub fn is_terminator(&self) -> bool {
        self.is_jump() || self.is_partition_exit() || self.is_return()
    }

    pub fn try_thread_id(&self) -> Option<(usize, ForkJoinID)> {
        if let SInst::ThreadID {
            dimension,
            fork_join,
        } = self
        {
            Some((*dimension, *fork_join))
        } else {
            None
        }
    }

    pub fn try_reduction_variable(&self) -> Option<usize> {
        if let SInst::ReductionVariable { number } = self {
            Some(*number)
        } else {
            None
        }
    }

    pub fn try_jump(&self) -> Option<(BlockID, Option<&ParallelEntry>, Option<&ReduceExit>)> {
        if let SInst::Jump {
            target,
            parallel_entry,
            reduce_exit,
        } = self
        {
            Some((*target, parallel_entry.as_ref(), reduce_exit.as_ref()))
        } else {
            None
        }
    }

    pub fn block_successors(&self) -> BlockSuccessors {
        match self {
            SInst::Jump {
                target,
                parallel_entry: _,
                reduce_exit: _,
            } => BlockSuccessors::One([*target]),
            SInst::Branch {
                cond: _,
                false_target,
                true_target,
            } => BlockSuccessors::Two([*false_target, *true_target]),
            _ => BlockSuccessors::Zero,
        }
    }

    pub fn upper_case_name(&self) -> &'static str {
        match self {
            SInst::Phi { inputs: _ } => "Phi",
            SInst::ThreadID {
                dimension: _,
                fork_join: _,
            } => "ThreadID",
            SInst::ReductionVariable { number: _ } => "ReductionVariable",
            SInst::Jump {
                target: _,
                parallel_entry: _,
                reduce_exit: _,
            } => "Jump",
            SInst::Branch {
                cond: _,
                false_target: _,
                true_target: _,
            } => "Branch",
            SInst::PartitionExit { data_outputs: _ } => "PartitionExit",
            SInst::Return { value: _ } => "Return",
            SInst::Unary { input: _, op } => op.upper_case_name(),
            SInst::Binary {
                left: _,
                right: _,
                op,
            } => op.upper_case_name(),
            SInst::Ternary {
                first: _,
                second: _,
                third: _,
                op,
            } => op.upper_case_name(),
            SInst::ProductExtract {
                product: _,
                indices: _,
            } => "ProductExtract",
            SInst::ProductInsert {
                product: _,
                data: _,
                indices: _,
            } => "ProductInsert",
            SInst::ArrayLoad {
                array: _,
                position: _,
                bounds: _,
            } => "ArrayLoad",
            SInst::ArrayStore {
                array: _,
                value: _,
                position: _,
                bounds: _,
            } => "ArrayStore",
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub enum BlockSuccessors {
    Zero,
    One([BlockID; 1]),
    Two([BlockID; 2]),
}

impl AsRef<[BlockID]> for BlockSuccessors {
    fn as_ref(&self) -> &[BlockID] {
        match self {
            BlockSuccessors::Zero => &[],
            BlockSuccessors::One(x) => x,
            BlockSuccessors::Two(x) => x,
        }
    }
}

/*
 * On entering a parallel section, we need to specify how many threads to spawn
 * and what the initial values of the reduction variables are.
 */
#[derive(Debug, Clone)]
pub struct ParallelEntry {
    pub thread_counts: Box<[SValue]>,
    pub reduce_inits: Box<[SValue]>,
}

/*
 * On exiting a reduce section, we need to specify which instructions in the
 * reduce block correspond to what reduction variables. This also specifies
 * which values defined inside the reduce block can be used outside the block.
 */
#[derive(Debug, Clone)]
pub struct ReduceExit {
    pub reduce_reducts: Box<[SValue]>,
}

/*
 * The operator types are mostly the same.
 */
#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum SUnaryOperator {
    Not,
    Neg,
    Cast(SType),
}

impl SUnaryOperator {
    pub fn upper_case_name(&self) -> &'static str {
        match self {
            SUnaryOperator::Not => "Not",
            SUnaryOperator::Neg => "Neg",
            SUnaryOperator::Cast(_) => "Cast",
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum SBinaryOperator {
    Add,
    Sub,
    Mul,
    Div,
    Rem,
    LT,
    LTE,
    GT,
    GTE,
    EQ,
    NE,
    Or,
    And,
    Xor,
    LSh,
    RSh,
}

impl SBinaryOperator {
    pub fn upper_case_name(&self) -> &'static str {
        match self {
            SBinaryOperator::Add => "Add",
            SBinaryOperator::Sub => "Sub",
            SBinaryOperator::Mul => "Mul",
            SBinaryOperator::Div => "Div",
            SBinaryOperator::Rem => "Rem",
            SBinaryOperator::LT => "LT",
            SBinaryOperator::LTE => "LTE",
            SBinaryOperator::GT => "GT",
            SBinaryOperator::GTE => "GTE",
            SBinaryOperator::EQ => "EQ",
            SBinaryOperator::NE => "NE",
            SBinaryOperator::Or => "Or",
            SBinaryOperator::And => "And",
            SBinaryOperator::Xor => "Xor",
            SBinaryOperator::LSh => "LSh",
            SBinaryOperator::RSh => "RSh",
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum STernaryOperator {
    Select,
}

impl STernaryOperator {
    pub fn upper_case_name(&self) -> &'static str {
        match self {
            STernaryOperator::Select => "Select",
        }
    }
}

pub type SFunctionName = String;

define_id_type!(ArrayID);
define_id_type!(BlockID);
define_id_type!(ForkJoinID);

define_dual_id_type!(InstID);
