extern crate hercules_ir;

use std::collections::{HashMap, HashSet, VecDeque};
use std::iter::{empty, once, zip};

use self::hercules_ir::*;

use crate::*;

/*
 * Iterate over the uses of a instruction.
 */
pub fn sched_get_uses(inst: &SInst) -> Box<dyn Iterator<Item = &SValue> + '_> {
    match inst {
        SInst::Phi { inputs } => Box::new(inputs.iter().map(|(_, svalue)| svalue)),
        SInst::ThreadID {
            dimension: _,
            fork_join: _,
        } => Box::new(empty()),
        SInst::ReductionVariable { number: _ } => Box::new(empty()),
        SInst::Jump {
            target: _,
            parallel_entry,
            reduce_exit,
        } => {
            let first = parallel_entry.as_ref().map(|parallel_entry| {
                parallel_entry
                    .thread_counts
                    .iter()
                    .chain(parallel_entry.reduce_inits.iter())
            });
            let second = reduce_exit
                .as_ref()
                .map(|reduce_exit| reduce_exit.reduce_reducts.iter());
            match (first, second) {
                (Some(first), Some(second)) => Box::new(first.chain(second)),
                (Some(first), None) => Box::new(first),
                (None, Some(second)) => Box::new(second),
                (None, None) => Box::new(empty()),
            }
        }
        SInst::Branch {
            cond,
            false_target: _,
            true_target: _,
        } => Box::new(once(cond)),
        SInst::PartitionExit { data_outputs } => Box::new(data_outputs.iter().map(|svalue| svalue)),
        SInst::Return { value } => Box::new(once(value)),
        SInst::Unary { input, op: _ } => Box::new(once(input)),
        SInst::Binary { left, right, op: _ } => Box::new(once(left).chain(once(right))),
        SInst::Ternary {
            first,
            second,
            third,
            op: _,
        } => Box::new(once(first).chain(once(second)).chain(once(third))),
        SInst::ProductExtract {
            product,
            indices: _,
        } => Box::new(once(product)),
        SInst::ProductInsert {
            product,
            data,
            indices: _,
        } => Box::new(once(product).chain(once(data))),
        SInst::ArrayLoad {
            array,
            position,
            bounds,
        } => Box::new(once(array).chain(position.iter()).chain(bounds.iter())),
        SInst::ArrayStore {
            array,
            value,
            position,
            bounds,
        } => Box::new(
            once(array)
                .chain(once(value))
                .chain(position.iter())
                .chain(bounds.iter()),
        ),
    }
}

/*
 * Map virtual registers to corresponding instruction IDs.
 */
pub fn sched_virt_reg_to_inst_id(function: &SFunction) -> HashMap<usize, InstID> {
    let mut virt_reg_to_inst_id = HashMap::new();
    for block_idx in 0..function.blocks.len() {
        let block = &function.blocks[block_idx];
        for inst_idx in 0..block.insts.len() {
            let virt_reg = block.virt_regs[inst_idx].0;
            let inst_id = InstID::new(block_idx, inst_idx);
            virt_reg_to_inst_id.insert(virt_reg, inst_id);
        }
    }
    virt_reg_to_inst_id
}

/*
 * Build a dependency graph of instructions in an SFunction.
 */
pub fn sched_dependence_graph(
    function: &SFunction,
    virt_reg_to_inst_id: &HashMap<usize, InstID>,
) -> HashMap<InstID, Vec<InstID>> {
    let mut dep_graph = HashMap::new();
    for inst_id in virt_reg_to_inst_id.values() {
        dep_graph.insert(*inst_id, vec![]);
    }

    // Process the dependencies in each block. This includes inter-block
    // dependencies for normal def-use edges.
    for block_idx in 0..function.blocks.len() {
        let block = &function.blocks[block_idx];

        // Add normal dependencies.
        for inst_idx in 0..block.insts.len() {
            let inst_id = InstID::new(block_idx, inst_idx);
            let inst = &block.insts[inst_idx];
            for use_sval in sched_get_uses(inst) {
                if let SValue::VirtualRegister(virt_reg) = use_sval {
                    // Uses of parameters don't correspond to any instruction we
                    // need to depend on.
                    if let Some(use_id) = virt_reg_to_inst_id.get(virt_reg) {
                        let deps = dep_graph.get_mut(&inst_id).unwrap();
                        if !deps.contains(use_id) {
                            deps.push(*use_id);
                        }
                    }
                }
            }
        }

        // Phis should appear at the top of linear basic blocks.

        // Add dependencies between the phis.
        let mut last_phi = None;
        for inst_idx in 0..block.insts.len() {
            let inst_id = InstID::new(block_idx, inst_idx);
            let inst = &block.insts[inst_idx];
            if inst.is_phi() {
                if let Some(last_phi) = last_phi {
                    let deps = dep_graph.get_mut(&inst_id).unwrap();
                    if !deps.contains(&last_phi) {
                        deps.push(last_phi);
                    }
                }
                last_phi = Some(inst_id);
            }
        }

        // If there is at least one phi, add a dependency between the "last" phi
        // and every non-phi instruction with no dependencies yet.
        if let Some(last_phi) = last_phi {
            for inst_idx in 0..block.insts.len() {
                let inst_id = InstID::new(block_idx, inst_idx);
                let inst = &block.insts[inst_idx];
                if !inst.is_phi() {
                    let deps = dep_graph.get_mut(&inst_id).unwrap();
                    if deps.is_empty() {
                        deps.push(last_phi);
                    }
                }
            }
        }

        // Terminator instructions appear at the bottom of linear basic blocks.

        // Find every non-terminator instruction with no users.
        let mut no_user_insts = (0..block.insts.len())
            .filter(|inst_idx| !block.insts[*inst_idx].is_terminator())
            .map(|inst_idx| InstID::new(block_idx, inst_idx))
            .collect::<HashSet<_>>();
        for inst_idx in 0..block.insts.len() {
            let inst_id = InstID::new(block_idx, inst_idx);
            for dep in dep_graph[&inst_id].iter() {
                no_user_insts.remove(dep);
            }
        }

        // Add a dependency between each instruction with no users (previously)
        // and each terminator instruction.
        for inst_idx in 0..block.insts.len() {
            let inst_id = InstID::new(block_idx, inst_idx);
            let inst = &block.insts[inst_idx];
            if inst.is_terminator() {
                let deps = dep_graph.get_mut(&inst_id).unwrap();
                for no_user_inst in no_user_insts.iter() {
                    if !deps.contains(no_user_inst) {
                        deps.push(*no_user_inst);
                    }
                }
            }
        }
    }

    dep_graph
}

/*
 * Assemble a map from SValue to SType.
 */
pub fn sched_svalue_types(function: &SFunction) -> HashMap<SValue, SType> {
    let mut result = HashMap::new();

    // Add types of parameters.
    for (param_idx, param_ty) in function.param_types.iter().enumerate() {
        result.insert(SValue::VirtualRegister(param_idx), param_ty.clone());
    }

    // Add types of instructions and constants.
    for block in function.blocks.iter() {
        for (inst, (virt_reg, sty)) in zip(block.insts.iter(), block.virt_regs.iter()) {
            // Add the type of the output of the instruction.
            result.insert(SValue::VirtualRegister(*virt_reg), sty.clone());

            // Find constants inthe uses of instructions.
            for u in sched_get_uses(inst) {
                if let SValue::Constant(cons) = u {
                    result.insert(u.clone(), cons.get_type());
                }
            }
        }
    }

    result
}

/*
 * Analysis information for one fork-join.
 */
#[derive(Debug)]
pub struct ParallelReduceInfo {
    // The block that jumps into the parallel section.
    pub predecessor: BlockID,
    // The block that is jumped into after the reduce section.
    pub successor: BlockID,

    // The first parallel block in the parallel section.
    pub top_parallel_block: BlockID,
    // The parallel block that jumps to the reduce section.
    pub bottom_parallel_block: BlockID,
    // The single block in the reduce section.
    pub reduce_block: BlockID,

    // The thread count SValues used for this fork-join.
    pub thread_counts: Box<[SValue]>,
    // The initial SValues for the reduction variables.
    pub reduce_inits: Box<[SValue]>,
    // The reduct SValues for the reduction variables.
    pub reduce_reducts: Box<[SValue]>,

    // Map from thread ID dimension to virtual registers of corresponding thread
    // ID instructions.
    pub thread_ids: HashMap<usize, Vec<usize>>,
    // Map from reduction variable number to virtual register of the
    // corresponding reduction variable instruction.
    pub reduction_variables: HashMap<usize, usize>,

    // If this parallel-reduce section is inside another parallel-reduce, store
    // the parent's ForkJoinID. Parallel-reduce sections in an SFunction form a
    // forest.
    pub parent_fork_join_id: Option<ForkJoinID>,

    // Information about how this fork-join should be scheduled. Collecting this
    // info here just makes writing the backends more convenient.
    pub vector_width: Option<usize>,
    // For each reduction variable, track if its associative or parallel
    // individually.
    pub associative_reduce: HashMap<usize, bool>,
    pub parallel_reduce: HashMap<usize, bool>,
    // Track if this is a "top-level" parallel-reduce. That is, the parallel-
    // reduce is the "only thing" inside this partition function. Only these
    // parallel-reduces can be parallelized on the CPU, even if this parallel-
    // reduce has a parallel schedule on the entry jump.
    pub top_level: bool,
}

/*
 * Analyze parallel-reduce sections to make lowering them easier. Returns a map
 * from ForkJoinID to information about that parallel-reduce section.
 */
pub fn sched_parallel_reduce_sections(
    function: &SFunction,
) -> HashMap<ForkJoinID, ParallelReduceInfo> {
    let mut result = HashMap::new();

    for (block_idx, block) in function.blocks.iter().enumerate() {
        // Start by identifying a jump into a parallel section.
        for (inst_idx, inst) in block.insts.iter().enumerate() {
            if let SInst::Jump {
                target,
                parallel_entry,
                reduce_exit: _,
            } = inst
                && let Some(parallel_entry) = parallel_entry
            {
                let predecessor = BlockID::new(block_idx);
                let ParallelEntry {
                    thread_counts,
                    reduce_inits,
                } = parallel_entry.clone();
                let vector_width = block.schedules[&inst_idx]
                    .iter()
                    .filter_map(|schedule| schedule.try_vectorizable())
                    .next();

                // The jump target is the top of the parallel section. Get the
                // fork-join ID from that block.
                let top_parallel_block = *target;
                let fork_join_id = function.blocks[top_parallel_block.idx()]
                    .kind
                    .try_parallel()
                    .unwrap();

                // Traverse the blocks until finding a jump to the corresponding
                // reduce block.
                let mut queue = VecDeque::from(vec![top_parallel_block]);
                let mut visited = HashSet::new();
                visited.insert(top_parallel_block);
                let mut bfs_dest = None;
                while let Some(bfs) = queue.pop_front() {
                    for succ in function.blocks[bfs.idx()].successors().as_ref() {
                        if let Some(reduce_fork_join_id) =
                            function.blocks[succ.idx()].kind.try_reduce()
                            && reduce_fork_join_id == fork_join_id
                        {
                            bfs_dest = Some((bfs, *succ));
                        } else if !visited.contains(succ) {
                            queue.push_back(*succ);
                            visited.insert(*succ);
                        }
                    }
                }
                let (bottom_parallel_block, reduce_block) = bfs_dest.unwrap();

                // Find the jump out of the reduce block.
                let (successor, _, reduce_exit) = function.blocks[reduce_block.idx()]
                    .insts
                    .iter()
                    .filter_map(|inst| inst.try_jump())
                    .next()
                    .unwrap();
                let reduce_reducts = reduce_exit.unwrap().reduce_reducts.clone();

                // Find the thread ID instructions.
                let mut thread_ids = (0..thread_counts.len())
                    .map(|dim| (dim, vec![]))
                    .collect::<HashMap<usize, Vec<usize>>>();
                for parallel_block in visited {
                    for (inst, (virt_reg, _)) in zip(
                        function.blocks[parallel_block.idx()].insts.iter(),
                        function.blocks[parallel_block.idx()].virt_regs.iter(),
                    ) {
                        if let Some((dim, tid_fork_join)) = inst.try_thread_id()
                            && tid_fork_join == fork_join_id
                        {
                            thread_ids.get_mut(&dim).unwrap().push(*virt_reg);
                        }
                    }
                }

                // Find the reduction variable instructions.
                let mut associative_reduce = HashMap::new();
                let mut parallel_reduce = HashMap::new();
                let reduce_sblock = &function.blocks[reduce_block.idx()];
                let reduction_variables = zip(
                    reduce_sblock.insts.iter().enumerate(),
                    reduce_sblock.virt_regs.iter(),
                )
                .filter_map(|((inst_idx, inst), (virt_reg, _))| {
                    inst.try_reduction_variable().map(|number| {
                        let schedules = &reduce_sblock.schedules[&inst_idx];
                        associative_reduce
                            .insert(number, schedules.contains(&SSchedule::Associative));
                        parallel_reduce
                            .insert(number, schedules.contains(&SSchedule::ParallelReduce));
                        (number, *virt_reg)
                    })
                })
                .collect();

                // Assemble all of the info and add it to the map.
                let info = ParallelReduceInfo {
                    predecessor,
                    successor,

                    top_parallel_block,
                    bottom_parallel_block,
                    reduce_block,

                    thread_counts,
                    reduce_inits,
                    reduce_reducts,

                    thread_ids,
                    reduction_variables,

                    parent_fork_join_id: None,
                    vector_width,
                    associative_reduce,
                    parallel_reduce,

                    top_level: false,
                };
                result.insert(fork_join_id, info);
            }
        }
    }

    // Figure out if any parallel-reduces are top level - that is, they are the
    // "only thing" in the partition function.
    for (_, parallel_reduce_info) in result.iter_mut() {
        // A parallel-reduce is top-level if its predecessor is the entry block
        // containing only a jump and its successor is an exit block containing
        // just a function terminator.
        let pred_block = &function.blocks[parallel_reduce_info.predecessor.idx()];
        let succ_block = &function.blocks[parallel_reduce_info.successor.idx()];
        if parallel_reduce_info.predecessor == BlockID::new(0)
            && pred_block.insts.len() == 1
            && pred_block.insts[0].is_jump()
            && succ_block.insts.len() == 1
            && (succ_block.insts[0].is_partition_exit() || succ_block.insts[0].is_return())
        {
            parallel_reduce_info.top_level = true;
        }
    }

    // Compute the parallel-reduce forest last, since this requires some info we
    // just computed above.
    let mut parents = HashMap::new();
    for (fork_join_id, parallel_reduce_info) in result.iter() {
        let mut pred_block = parallel_reduce_info.predecessor;

        // Keep looking at predecessors of adjacent parallel-reduce sections
        // until one belongs to a parent parallel-reduce or is sequential, so
        // this parallel-reduce is a root.
        let parent = loop {
            match function.blocks[pred_block.idx()].kind {
                // If the predecessor is sequential, then this parallel-reduce
                // is a root.
                SBlockKind::Sequential => break None,
                // If the predecessor is parallel, then this parallel-reduce is
                // inside that parallel-reduce.
                SBlockKind::Parallel(parent) => break Some(parent),
                // If the predecessor is reduce, then that parallel-reduce is a
                // child of the same parent. Iterate on its predecessor.
                SBlockKind::Reduce(adjacent) => {
                    pred_block = result[&adjacent].predecessor;
                }
            }
        };
        parents.insert(*fork_join_id, parent);
    }

    // Insert the information into the parallel reduce info map.
    for (fork_join_id, parallel_reduce_info) in result.iter_mut() {
        parallel_reduce_info.parent_fork_join_id = parents[fork_join_id];
    }

    result
}
