use std::collections::{HashMap, VecDeque};
use std::iter::{repeat, zip};

use crate::*;

/*
 * An individual schedule is a single "directive" for the compiler to take into
 * consideration at some point during the compilation pipeline. Each schedule is
 * associated with a single node.
 */
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Schedule {
    // This fork can be run in parallel and has a "natural" tiling, which may or
    // may not be respected by certain backends. The field stores at least how
    // many parallel tiles should be run concurrently, along each dimension.
    // Some backends (such as GPU) may spawn more parallel tiles (each tile
    // being a single thread in that case) along each axis.
    ParallelFork(Box<[usize]>),
    // This reduce can be "run in parallel" - conceptually, the `reduct`
    // backedge can be removed, and the reduce code can be merged into the
    // parallel code.
    ParallelReduce,
    // This fork-join has no impeding control flow. The field stores the vector
    // width.
    Vectorizable(usize),
    // This reduce can be re-associated. This may lower a sequential dependency
    // chain into a reduction tree.
    Associative,
}

/*
 * The authoritative enumeration of supported devices. Technically, a device
 * refers to a specific backend, so different "devices" may refer to the same
 * "kind" of hardware.
 */
#[derive(Debug, Clone, Copy)]
pub enum Device {
    CPU,
    GPU,
    // Hercules function calls are placed in solitary function calls that are
    // directly represented in the generated async Rust runtime code.
    AsyncRust,
}

/*
 * A plan is a set of schedules associated with each node, plus partitioning
 * information. Partitioning information is a mapping from node ID to partition
 * ID. A plan's scope is a single function.
 */
#[derive(Debug, Clone)]
pub struct Plan {
    pub schedules: Vec<Vec<Schedule>>,
    pub partitions: Vec<PartitionID>,
    pub partition_devices: Vec<Device>,
    pub num_partitions: usize,
}

define_id_type!(PartitionID);

impl Plan {
    /*
     * Invert stored map from node to partition to map from partition to nodes.
     */
    pub fn invert_partition_map(&self) -> Vec<Vec<NodeID>> {
        let mut map = vec![vec![]; self.num_partitions];

        for idx in 0..self.partitions.len() {
            map[self.partitions[idx].idx()].push(NodeID::new(idx));
        }

        map
    }

    /*
     * GCM takes in a partial partitioning, but most of the time we have a full
     * partitioning.
     */
    pub fn make_partial_partitioning(&self) -> Vec<Option<PartitionID>> {
        self.partitions.iter().map(|id| Some(*id)).collect()
    }

    /*
     * LEGACY API: This is the legacy mechanism for repairing plans. Future code
     * should use the `repair_plan` function in editor.rs.
     * Plans must be "repairable", in the sense that the IR that's referred to
     * may change after many passes. Since a plan is an explicit side data
     * structure, it must be updated after every change in the IR.
     */
    pub fn repair(self, function: &Function, grave_mapping: &Vec<NodeID>) -> Self {
        // Unpack the plan.
        let old_inverse_partition_map = self.invert_partition_map();
        let Plan {
            mut schedules,
            partitions: _,
            partition_devices,
            num_partitions: _,
        } = self;

        // Schedules of old nodes just get dropped. Since schedules don't hold
        // necessary semantic information, we are free to drop them arbitrarily.
        schedules.fix_gravestones(grave_mapping);
        schedules.resize(function.nodes.len(), vec![]);

        // Delete now empty partitions. First, filter out deleted nodes from the
        // partitions and simultaneously map old node IDs to new node IDs. Then,
        // filter out empty partitions.
        let (new_inverse_partition_map, new_devices): (Vec<Vec<NodeID>>, Vec<Device>) =
            zip(old_inverse_partition_map, partition_devices)
                .into_iter()
                .map(|(contents, device)| {
                    (
                        contents
                            .into_iter()
                            .filter_map(|id| {
                                if id.idx() == 0 || grave_mapping[id.idx()].idx() != 0 {
                                    Some(grave_mapping[id.idx()])
                                } else {
                                    None
                                }
                            })
                            .collect::<Vec<NodeID>>(),
                        device,
                    )
                })
                .filter(|(contents, _)| !contents.is_empty())
                .unzip();

        // Calculate the number of nodes after deletion but before addition. Use
        // this is iterate new nodes later.
        let num_nodes_before_addition = new_inverse_partition_map.iter().flatten().count();
        assert!(new_inverse_partition_map
            .iter()
            .flatten()
            .all(|id| id.idx() < num_nodes_before_addition));

        // Calculate the nodes that need to be assigned to a partition. This
        // starts as just the nodes that have been added by passes.
        let mut new_node_ids: VecDeque<NodeID> = (num_nodes_before_addition..function.nodes.len())
            .map(NodeID::new)
            .collect();

        // Any partition no longer containing at least one control node needs to
        // be liquidated.
        let (new_inverse_partition_map, new_devices): (Vec<Vec<NodeID>>, Vec<Device>) =
            zip(new_inverse_partition_map, new_devices)
                .into_iter()
                .filter_map(|(part, device)| {
                    if part.iter().any(|id| function.nodes[id.idx()].is_control()) {
                        Some((part, device))
                    } else {
                        // Nodes in removed partitions need to be re-partitioned.
                        new_node_ids.extend(part);
                        None
                    }
                })
                .unzip();

        // Assign the node IDs that need to be partitioned to partitions. In the
        // process, construct a map from node ID to partition ID.
        let mut node_id_to_partition_id: HashMap<NodeID, PartitionID> = new_inverse_partition_map
            .into_iter()
            .enumerate()
            .map(|(partition_idx, node_ids)| {
                node_ids
                    .into_iter()
                    .map(|node_id| (node_id, PartitionID::new(partition_idx)))
                    .collect::<Vec<(NodeID, PartitionID)>>()
            })
            .flatten()
            .collect();

        // Make a best effort to assign nodes to the partition of one of their
        // uses. Prioritize earlier uses. TODO: since not all partitions are
        // legal, this is almost certainly not complete. Think more about that.
        'workloop: while let Some(id) = new_node_ids.pop_front() {
            for u in get_uses(&function.nodes[id.idx()]).as_ref() {
                if let Some(partition_id) = node_id_to_partition_id.get(u) {
                    node_id_to_partition_id.insert(id, *partition_id);
                    continue 'workloop;
                }
            }
            new_node_ids.push_back(id);
        }

        // Reconstruct the partitions vector.
        let num_partitions = new_devices.len();
        let mut partitions = vec![PartitionID::new(0); function.nodes.len()];
        for (k, v) in node_id_to_partition_id {
            partitions[k.idx()] = v;
        }

        // Reconstruct the whole plan.
        Plan {
            schedules,
            partitions,
            partition_devices: new_devices,
            num_partitions,
        }
    }

    /*
     * Verify that a partitioning is valid.
     */
    pub fn verify_partitioning(
        &self,
        function: &Function,
        def_use: &ImmutableDefUseMap,
        fork_join_map: &HashMap<NodeID, NodeID>,
    ) {
        let partition_to_node_ids = self.invert_partition_map();

        // First, verify that there is at most one control node in the partition
        // with a control use outside the partition. A partition may only have
        // zero such control nodes if it contains the start node. This also
        // checks that each partition has at least one control node.
        for nodes_in_partition in partition_to_node_ids.iter() {
            let contains_start = nodes_in_partition
                .iter()
                .any(|id| function.nodes[id.idx()] == Node::Start);
            let num_inter_partition_control_uses = nodes_in_partition
                .iter()
                .filter(|id| {
                    // An inter-partition control use is a control node,
                    function.nodes[id.idx()].is_control()
                        // where one of its uses,
                        && get_uses(&function.nodes[id.idx()])
                            .as_ref()
                            .into_iter()
                            .any(|use_id| {
                                // that is itself a control node as well,
                                function.nodes[use_id.idx()].is_control()
                                    // is in a different partition.
                                    && self.partitions[use_id.idx()] != self.partitions[id.idx()]
                            })
                })
                .count();

            assert!(num_inter_partition_control_uses + contains_start as usize == 1, "PANIC: Found an invalid partition based on the inter-partition control use criteria.");
        }

        // Second, verify that fork-joins are not split amongst partitions.
        for id in (0..function.nodes.len()).map(NodeID::new) {
            if function.nodes[id.idx()].is_fork() {
                let fork_part = self.partitions[id.idx()];

                // The join must be in the same partition.
                let join = fork_join_map[&id];
                assert_eq!(
                    fork_part,
                    self.partitions[join.idx()],
                    "PANIC: Join is in a different partition than its corresponding fork."
                );

                // The thread IDs must be in the same partition.
                def_use
                    .get_users(id)
                    .into_iter()
                    .filter(|user| function.nodes[user.idx()].is_thread_id())
                    .for_each(|thread_id| {
                        assert_eq!(
                            fork_part,
                            self.partitions[thread_id.idx()],
                            "PANIC: Thread ID is in a different partition than its fork use."
                        )
                    });

                // The reduces must be in the same partition.
                def_use
                    .get_users(join)
                    .into_iter()
                    .filter(|user| function.nodes[user.idx()].is_reduce())
                    .for_each(|reduce| {
                        assert_eq!(
                            fork_part,
                            self.partitions[reduce.idx()],
                            "PANIC: Reduce is in a different partition than its join use."
                        )
                    });
            }
        }

        // Third, verify that every data node has proper dominance relations
        // with respect to the partitioning. In particular:
        // 1. Every non-phi data node should be in a partition that is dominated
        //    by the partitions of every one of its uses.
        // 2. Every data node should be in a partition that dominates the
        //    partitions of every one of its non-phi users.
        // Compute a dominance relation between the partitions by constructing a
        // partition control graph.
        let partition_graph = partition_graph(function, def_use, self);
        let dom = dominator(&partition_graph, NodeID::new(self.partitions[0].idx()));
        for id in (0..function.nodes.len()).map(NodeID::new) {
            if function.nodes[id.idx()].is_control() {
                continue;
            }
            let part = self.partitions[id.idx()];

            // Check condition #1.
            if !function.nodes[id.idx()].is_phi() {
                let uses = get_uses(&function.nodes[id.idx()]);
                for use_id in uses.as_ref() {
                    let use_part = self.partitions[use_id.idx()];
                    assert!(dom.does_dom(NodeID::new(use_part.idx()), NodeID::new(part.idx())), "PANIC: A data node has a partition use that doesn't dominate its partition.");
                }
            }

            // Check condition #2.
            let users = def_use.get_users(id);
            for user_id in users.as_ref() {
                if !function.nodes[user_id.idx()].is_phi() {
                    let user_part = self.partitions[user_id.idx()];
                    assert!(dom.does_dom(NodeID::new(part.idx()), NodeID::new(user_part.idx())), "PANIC: A data node has a partition user that isn't dominated by its partition.");
                }
            }
        }

        // Fourth, verify that every projection node is in the same partition as
        // its control use.
        for id in (0..function.nodes.len()).map(NodeID::new) {
            if let Node::Projection {
                control,
                selection: _,
            } = function.nodes[id.idx()]
            {
                assert_eq!(
                    self.partitions[id.idx()],
                    self.partitions[control.idx()],
                    "PANIC: Found a projection node in a different partition than its control use."
                );
            }
        }

        // Fifth, verify that every partition has at least one partition
        // successor xor has at least one return node.
        for partition_idx in 0..self.num_partitions {
            let has_successor = partition_graph.succs(NodeID::new(partition_idx)).count() > 0;
            let has_return = partition_to_node_ids[partition_idx]
                .iter()
                .any(|node_id| function.nodes[node_id.idx()].is_return());
            assert!(has_successor ^ has_return, "PANIC: Found an invalid partition based on the partition return / control criteria.");
        }
    }

    /*
     * Compute the top node for each partition.
     */
    pub fn compute_top_nodes(
        &self,
        function: &Function,
        control_subgraph: &Subgraph,
        inverted_partition_map: &Vec<Vec<NodeID>>,
    ) -> Vec<NodeID> {
        inverted_partition_map
            .into_iter()
            .enumerate()
            .map(|(part_idx, part)| {
                // For each partition, find the "top" node.
                *part
                    .iter()
                    .filter(move |id| {
                        // The "top" node is a control node having at least one
                        // control predecessor in another partition, or is a
                        // start node. Every predecessor in the control subgraph
                        // is a control node.
                        function.nodes[id.idx()].is_start()
                            || (function.nodes[id.idx()].is_control()
                                && control_subgraph
                                    .preds(**id)
                                    .filter(|pred_id| {
                                        self.partitions[pred_id.idx()].idx() != part_idx
                                    })
                                    .count()
                                    > 0)
                    })
                    // We assume here there is exactly one such top node per
                    // partition. Verify a partitioning with
                    // `verify_partitioning` before calling this method.
                    .next()
                    .unwrap()
            })
            .collect()
    }

    /*
     * Compute the data inputs of each partition.
     */
    pub fn compute_data_inputs(&self, function: &Function) -> Vec<Vec<NodeID>> {
        let mut data_inputs = vec![vec![]; self.num_partitions];

        // First consider the non-phi nodes in each partition.
        for id in (0..function.nodes.len()).map(NodeID::new) {
            if function.nodes[id.idx()].is_phi() {
                continue;
            }

            let data_inputs = &mut data_inputs[self.partitions[id.idx()].idx()];
            let uses = get_uses(&function.nodes[id.idx()]);
            for use_id in uses.as_ref() {
                // For every non-phi node, check each of its data uses. If the
                // node and its use are in different partitions, then the use is
                // a data input for the partition of the node. Also, don't add
                // the same node to the data inputs list twice. Only consider
                // non-constant uses data inputs - constant nodes are always
                // rematerialized into the user partition.
                if !function.nodes[use_id.idx()].is_control()
                    && !function.nodes[use_id.idx()].is_constant()
                    && self.partitions[id.idx()] != self.partitions[use_id.idx()]
                    && !data_inputs.contains(use_id)
                {
                    data_inputs.push(*use_id);
                }
            }
        }

        // Second consider the phi nodes in each partition.
        for id in (0..function.nodes.len()).map(NodeID::new) {
            if !function.nodes[id.idx()].is_phi() {
                continue;
            }

            let data_inputs = &mut data_inputs[self.partitions[id.idx()].idx()];
            let uses = get_uses(&function.nodes[id.idx()]);
            for use_id in uses.as_ref() {
                // For every phi node, if any one of its non-constant uses is
                // defined in a different partition, then the phi node itself,
                // not its outside uses, is considered a data input. This is
                // because a phi node whose uses are all in a different
                // partition should be lowered to a single parameter to the
                // corresponding simple IR function. Note that for a phi node
                // with some uses outside and some uses inside the partition,
                // the uses outside the partition become a single parameter to
                // the schedule IR function, and that parameter and all of the
                // "inside" uses become the inputs to a phi inside the simple IR
                // function.
                if self.partitions[id.idx()] != self.partitions[use_id.idx()]
                    && !function.nodes[use_id.idx()].is_constant()
                    && !data_inputs.contains(&id)
                {
                    data_inputs.push(id);
                    break;
                }
            }
        }

        // Sort the node IDs to keep a consistent interface between partitions.
        for data_inputs in &mut data_inputs {
            data_inputs.sort();
        }
        data_inputs
    }

    /*
     * Compute the data outputs of each partition.
     */
    pub fn compute_data_outputs(
        &self,
        function: &Function,
        def_use: &ImmutableDefUseMap,
    ) -> Vec<Vec<NodeID>> {
        let mut data_outputs = vec![vec![]; self.num_partitions];

        for id in (0..function.nodes.len()).map(NodeID::new) {
            // Only consider non-constant data nodes as data outputs, since
            // constant nodes are rematerialized into the user partition.
            if function.nodes[id.idx()].is_control() || function.nodes[id.idx()].is_constant() {
                continue;
            }

            let data_outputs = &mut data_outputs[self.partitions[id.idx()].idx()];
            let users = def_use.get_users(id);
            for user_id in users.as_ref() {
                // For every data node, check each of its users. If the node and
                // its user are in different partitions, then the node is a data
                // output for the partition of the node. Also, don't add the
                // same node to the data outputs list twice. It doesn't matter
                // how this data node is being used - all that matters is that
                // it itself is a data node, and that it has a user outside the
                // partition. This makes the code simpler than the inputs case.
                if self.partitions[id.idx()] != self.partitions[user_id.idx()]
                    && !data_outputs.contains(&id)
                {
                    data_outputs.push(id);
                    break;
                }
            }
        }

        // Sort the node IDs to keep a consistent interface between partitions.
        for data_outputs in &mut data_outputs {
            data_outputs.sort();
        }
        data_outputs
    }

    pub fn renumber_partitions(&mut self) {
        let mut renumber_partitions = HashMap::new();
        for id in self.partitions.iter_mut() {
            let next_id = PartitionID::new(renumber_partitions.len());
            let old_id = *id;
            let new_id = *renumber_partitions.entry(old_id).or_insert(next_id);
            *id = new_id;
        }
        let mut new_devices = vec![Device::CPU; renumber_partitions.len()];
        for (old_id, new_id) in renumber_partitions.iter() {
            new_devices[new_id.idx()] = self.partition_devices[old_id.idx()];
        }
        self.partition_devices = new_devices;
        self.num_partitions = renumber_partitions.len();
    }
}

impl GraveUpdatable for Plan {
    fn fix_gravestones(&mut self, grave_mapping: &Vec<NodeID>) {
        self.schedules.fix_gravestones(grave_mapping);
        self.partitions.fix_gravestones(grave_mapping);
        self.renumber_partitions();
    }
}

/*
 * Infer parallel fork-joins. These are fork-joins with only parallel reduction
 * variables and no parent fork-joins.
 */
pub fn infer_parallel_fork(
    function: &Function,
    def_use: &ImmutableDefUseMap,
    fork_join_map: &HashMap<NodeID, NodeID>,
    fork_join_nest: &HashMap<NodeID, Vec<NodeID>>,
    plan: &mut Plan,
) {
    for id in (0..function.nodes.len()).map(NodeID::new) {
        let Node::Fork {
            control: _,
            ref factors,
        } = function.nodes[id.idx()]
        else {
            continue;
        };
        let join_id = fork_join_map[&id];
        let all_parallel_reduce = def_use.get_users(join_id).as_ref().into_iter().all(|user| {
            plan.schedules[user.idx()].contains(&Schedule::ParallelReduce)
                || function.nodes[user.idx()].is_control()
        });
        let top_level = fork_join_nest[&id].len() == 1 && fork_join_nest[&id][0] == id;
        if all_parallel_reduce && top_level {
            let tiling = repeat(4).take(factors.len()).collect();
            plan.schedules[id.idx()].push(Schedule::ParallelFork(tiling));
        }
    }
}

/*
 * Infer parallel reductions consisting of a simple cycle between a Reduce node
 * and a Write node, where indices of the Write are position indices using the
 * ThreadID nodes attached to the corresponding Fork. This procedure also adds
 * the ParallelReduce schedule to Reduce nodes reducing over a parallelized
 * Reduce, as long as the base Write node also has position indices of the
 * ThreadID of the outer fork. In other words, the complete Reduce chain is
 * annotated with ParallelReduce, as long as each ThreadID dimension appears in
 * the positional indexing of the original Write.
 */
pub fn infer_parallel_reduce(
    function: &Function,
    fork_join_map: &HashMap<NodeID, NodeID>,
    plan: &mut Plan,
) {
    for id in (0..function.nodes.len())
        .map(NodeID::new)
        .filter(|id| function.nodes[id.idx()].is_reduce())
    {
        let mut first_control = None;
        let mut last_reduce = id;
        let mut chain_id = id;

        // Walk down Reduce chain until we reach the Reduce potentially looping
        // with the Write. Note the control node of the first Reduce, since this
        // will tell us which Thread ID to look for in the Write.
        while let Node::Reduce {
            control,
            init: _,
            reduct,
        } = function.nodes[chain_id.idx()]
        {
            if first_control.is_none() {
                first_control = Some(control);
            }

            last_reduce = chain_id;
            chain_id = reduct;
        }

        // Check for a Write-Reduce tight cycle.
        if let Node::Write {
            collect,
            data: _,
            indices,
        } = &function.nodes[chain_id.idx()]
            && *collect == last_reduce
        {
            // If there is a Write-Reduce tight cycle, get the position indices.
            let positions = indices
                .iter()
                .filter_map(|index| {
                    if let Index::Position(indices) = index {
                        Some(indices)
                    } else {
                        None
                    }
                })
                .flat_map(|pos| pos.iter());

            // Get the Forks corresponding to uses of bare ThreadIDs.
            let fork_thread_id_pairs = positions.filter_map(|id| {
                if let Node::ThreadID { control, dimension } = function.nodes[id.idx()] {
                    Some((control, dimension))
                } else {
                    None
                }
            });
            let mut forks = HashMap::<NodeID, Vec<usize>>::new();
            for (fork, dim) in fork_thread_id_pairs {
                forks.entry(fork).or_default().push(dim);
            }

            // Check if one of the Forks correspond to the Join associated with
            // the Reduce being considered, and has all of its dimensions
            // represented in the indexing.
            let is_parallel = forks.into_iter().any(|(id, mut rep_dims)| {
                rep_dims.sort();
                rep_dims.dedup();
                fork_join_map[&id] == first_control.unwrap()
                    && function.nodes[id.idx()].try_fork().unwrap().1.len() == rep_dims.len()
            });

            if is_parallel {
                plan.schedules[id.idx()].push(Schedule::ParallelReduce);
            }
        }
    }
}

/*
 * Infer vectorizable fork-joins. Just check that there are no control nodes
 * between a fork and its join and the factor is a constant.
 */
pub fn infer_vectorizable(
    function: &Function,
    dynamic_constants: &Vec<DynamicConstant>,
    fork_join_map: &HashMap<NodeID, NodeID>,
    plan: &mut Plan,
) {
    for id in (0..function.nodes.len())
        .map(NodeID::new)
        .filter(|id| function.nodes[id.idx()].is_join())
    {
        let u = get_uses(&function.nodes[id.idx()]).as_ref()[0];
        if let Some(join) = fork_join_map.get(&u)
            && *join == id
        {
            let factors = function.nodes[u.idx()].try_fork().unwrap().1;
            if factors.len() == 1
                && let Some(width) = evaluate_dynamic_constant(factors[0], dynamic_constants)
            {
                plan.schedules[u.idx()].push(Schedule::Vectorizable(width));
            }
        }
    }
}

/*
 * Infer associative reduction loops.
 */
pub fn infer_associative(function: &Function, plan: &mut Plan) {
    let is_associative = |op| match op {
        BinaryOperator::Add
        | BinaryOperator::Mul
        | BinaryOperator::Or
        | BinaryOperator::And
        | BinaryOperator::Xor => true,
        _ => false,
    };

    for (id, reduct) in (0..function.nodes.len()).map(NodeID::new).filter_map(|id| {
        function.nodes[id.idx()]
            .try_reduce()
            .map(|(_, _, reduct)| (id, reduct))
    }) {
        if let Node::Binary { left, right, op } = function.nodes[reduct.idx()]
            && (left == id || right == id)
            && is_associative(op)
        {
            plan.schedules[id.idx()].push(Schedule::Associative);
        }
    }
}

/*
 * Create partitions corresponding to fork-join nests. Also, split the "top-
 * level" partition into sub-partitions that are connected graphs. Place data
 * nodes using ThreadID or Reduce nodes in the corresponding fork-join nest's
 * partition.
 */
pub fn partition_out_forks(
    function: &Function,
    reverse_postorder: &Vec<NodeID>,
    fork_join_map: &HashMap<NodeID, NodeID>,
    bbs: &Vec<NodeID>,
    plan: &mut Plan,
) {
    #[allow(non_local_definitions)]
    impl Semilattice for NodeID {
        fn meet(a: &Self, b: &Self) -> Self {
            if a.idx() < b.idx() {
                *a
            } else {
                *b
            }
        }

        fn bottom() -> Self {
            NodeID::new(0)
        }

        fn top() -> Self {
            NodeID::new(!0)
        }
    }

    // Step 1: do dataflow analysis over control nodes to identify a
    // representative node for each partition. Each fork not taking as input the
    // ID of another fork node introduces its own ID as a partition
    // representative. Each join node propagates the fork ID if it's not the
    // fork pointing to the join in the fork join map - otherwise, it introduces
    // its user as a representative node ID for a partition. A region node taking
    // multiple node IDs as input belongs to the partition with the smaller
    // representative node ID.
    let mut representatives = forward_dataflow(
        function,
        reverse_postorder,
        |inputs: &[&NodeID], node_id: NodeID| match function.nodes[node_id.idx()] {
            Node::Start => NodeID::new(0),
            Node::Fork {
                control,
                factors: _,
            } => {
                // Start a partition if the preceding partition isn't a fork
                // partition and the predecessor isn't the join for the
                // predecessor fork partition. Otherwise, be part of the parent
                // fork partition.
                if *inputs[0] != NodeID::top()
                    && function.nodes[inputs[0].idx()].is_fork()
                    && fork_join_map.get(&inputs[0]) != Some(&control)
                {
                    inputs[0].clone()
                } else {
                    node_id
                }
            }
            Node::Join { control: _ } => inputs[0].clone(),
            _ => {
                // If the previous node is a join and terminates a fork's
                // partition, then start a new partition here. Otherwise, just
                // meet over the input lattice values. Set all data nodes to be
                // in the !0 partition.
                if !function.nodes[node_id.idx()].is_control() {
                    NodeID::top()
                } else if zip(inputs, get_uses(&function.nodes[node_id.idx()]).as_ref())
                    .any(|(part_id, pred_id)| fork_join_map.get(part_id) == Some(pred_id))
                {
                    node_id
                } else {
                    inputs
                        .iter()
                        .filter(|id| {
                            ***id != NodeID::top() && function.nodes[id.idx()].is_control()
                        })
                        .fold(NodeID::top(), |a, b| NodeID::meet(&a, b))
                }
            }
        },
    );

    // Step 2: assign data nodes to the partitions of the control nodes they are
    // assigned to by GCM.
    for idx in 0..function.nodes.len() {
        if !function.nodes[idx].is_control() {
            representatives[idx] = representatives[bbs[idx].idx()];
        }
    }

    // Step 3: deduplicate representative node IDs.
    let mut representative_to_partition_ids = HashMap::new();
    for rep in &representatives {
        if !representative_to_partition_ids.contains_key(rep) {
            representative_to_partition_ids
                .insert(rep, PartitionID::new(representative_to_partition_ids.len()));
        }
    }

    // Step 4: update plan.
    plan.num_partitions = representative_to_partition_ids.len();
    for id in (0..function.nodes.len()).map(NodeID::new) {
        plan.partitions[id.idx()] = representative_to_partition_ids[&representatives[id.idx()]];
    }

    plan.partition_devices = vec![Device::CPU; plan.num_partitions];
}

/*
 * Set the device for all partitions containing a fork to the GPU.
 */
pub fn place_fork_partitions_on_gpu(function: &Function, plan: &mut Plan) {
    for idx in 0..function.nodes.len() {
        if function.nodes[idx].is_fork() {
            plan.partition_devices[plan.partitions[idx].idx()] = Device::GPU;
        }
    }
}
